import torch
from typing import List, Union
import re
import subprocess


def is_available():
    return torch.cuda.is_available()


def device_count():
    if not is_available():
        return 0

    return torch.cuda.device_count()


def all_memory_cached(device_id: Union[List[int], int] = None, process=False):
    from thexp.base_classes import llist
    match_mem = re.compile('([0-9]+[a-zA-Z]+) \/ ([0-9]+[a-zA-Z]+)')
    proc = subprocess.Popen(['nvidia-smi'], stdout=subprocess.PIPE)

    count = device_count()
    res = llist()
    for i, line in enumerate(proc.stdout.readlines()):
        if i < 7:
            continue

        if i % 3 == 2:
            line = line.decode('utf-8').strip()
            res_ = re.findall(match_mem, line)
            res.append(res_[1][0])
            count -= 1
        if count == 0:
            break


    if device_id is not None:
        res = res[device_id]


    if process:
        # TODO 待验证
        match_num = re.compile('([0-9]+)')
        for i in range(len(res)):
            res[i] = re.findall(match_num,res[i])

    return res



import numpy as np
import torch


class RandomCrop():
    def __init__(self, pad=200):
        self.pad = pad

    def __call__(self, x):
        size = x.shape[0]
        x = np.pad(x, self.pad, mode='wrap')
        n = np.random.randint(0, self.pad)
        return x[n:n + size]


class RandomShift():
    def __call__(self, x):
        return np.roll(x, np.random.randint(0, x.shape[0] // 2), axis=-1)


class RandomNoisy():

    def __init__(self, rate=0.00001) -> None:
        super().__init__()
        self.rate = rate

    def __call__(self, x):
        wn = np.random.randn(*x.shape)
        return x + np.random.rand() * self.rate * wn


class RandomGain():
    def __init__(self, low=0.8, high=1.2):
        self.low = low
        self.high = high

    def __call__(self, x):
        return np.clip(x * (self.low + np.random.rand() * self.high), a_min=-32768, a_max=32767)


class PadTo():
    def __init__(self, to_size=20320):
        self.to_size = to_size

    def __call__(self, x):
        left = (self.to_size - x.shape[0]) // 2
        right = (self.to_size - x.shape[0]) - left
        return np.pad(x, (left, right), mode='constant', constant_values=0)


class ToTensor():
    def __call__(self, x):
        return torch.tensor(x, dtype=torch.float)



"""
Methods about files/paths/hash
"""
import hashlib
import json
import os
from functools import lru_cache

from ..globals import _GITKEY


def listdir_by_time(dir_path):
    dir_list = os.listdir(dir_path)
    if not dir_list:
        return []
    else:
        # os.path.getctime() 函数是获取文件最后创建时间
        dir_list = sorted(dir_list, key=lambda x: os.path.getatime(os.path.join(dir_path, x)), reverse=True)
        return dir_list


def _default_config():
    return {
        _GITKEY.expsdir: os.path.expanduser("~/.thexp/experiments")
    }


def _create_home_dir(path):
    os.makedirs(path)
    import json
    with open("config.json", "w") as w:
        json.dump({
            _GITKEY.expsdir: os.path.join(path, 'experiments')
        }, w, indent=2)


@lru_cache(1)
def home_dir():
    """dir for storing global record"""
    path = os.path.expanduser("~/.thexp")
    if not os.path.exists(path):
        _create_home_dir(path)

    return path


@lru_cache(1)
def config_path():
    """global config file path"""
    return os.path.join(home_dir(), "config.json")


def write_global_config(dumpsrc: dict):
    """
    write global config
    Notes:
    ------
    User should not call this method directly
    """
    path = config_path()
    with open(path, "w") as w:
        return json.dump(dumpsrc, w, indent=2)


def global_config() -> dict:
    """load global config"""
    path = config_path()
    if not os.path.exists(path):
        write_global_config(_default_config())
    with open(path, "r") as r:
        return json.load(r)


def file_atime_hash(file):
    """
    calculate hash of given file's atime
    atime : time of last access
    """
    return string_hash(str(os.path.getatime(file)))


def string_hash(*str):
    """calculate hash of given string list"""
    hl = hashlib.md5()
    for s in str:
        hl.update(s.encode(encoding='utf-8'))
    return hl.hexdigest()[:16]


def file_hash(file: str) -> str:
    """calculate hash of given file path"""
    hl = hashlib.md5()
    with open(file, 'rb') as r:
        hl.update(r.read())
    return hl.hexdigest()[:16]


def filter_filename(title: str, substr='-'):
    """replace invalid string of given file path by `substr`"""
    import re
    title = re.sub('[\/:*?"<>|]', substr, title)  # 去掉非法字符
    return title


def hash(value) -> str:
    """try to calculate hash of any given object"""
    import hashlib
    from collections.abc import Iterable
    from numbers import Number
    from numpy import ndarray
    from torch import Tensor
    hl = hashlib.md5()

    if isinstance(value, (ndarray, Tensor)):
        if isinstance(hl, Tensor):
            value = value.detach_().cpu().numpy()
        try:
            value = value.item()
        except:
            value = None
    if isinstance(value, (Number)):
        value = str(value)

    if isinstance(value, dict):
        for k in sorted(value.keys()):
            v = value[k]
            hl.update(str(k).encode(encoding='utf-8'))
            hl.update(hash(v).encode(encoding='utf-8'))
    elif isinstance(value, str):
        hl.update(value.encode(encoding='utf-8'))
    elif isinstance(value, Iterable):
        for v in value:
            hl.update(hash(v).encode(encoding='utf-8'))
    return hl.hexdigest()


def renormpath(path):
    return os.path.normcase(path).replace("\\", '/')



"""

"""
import torch
from torch import nn
from thexp import Params, Trainer, callbacks, Meter


class MyModel(nn.Module):
    def __init__(self):
        super().__init__()
        self.fc = nn.Linear(28 * 28, 10)

    def forward(self, x):
        x = x.view(x.shape[0], -1)
        x = self.fc(x)
        return x


class TCallbacks(callbacks.TrainCallback):
    def __init__(self):
        self.mark = set()

    def on_initial_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        super().on_initial_end(trainer, func, params, meter, *args, **kwargs)
        self.mark.add(1)


class MyTrainer(Trainer):

    def callbacks(self, params: Params):
        super().callbacks(params)
        TCallbacks().hook(self)

    def models(self, params: Params):
        super().models(params)
        self.rnd.mark('test')
        self.modela = MyModel()
        self.rnd.mark('test')
        self.modelb = MyModel()
        for pa, pb in zip(self.modela.parameters(), self.modelb.parameters()):
            assert (pa.data == pb.data).all()

    def train_batch(self, eidx, idx, global_step, batch_data, params: Params, device: torch.device):
        super().train_batch(eidx, idx, global_step, batch_data, params, device)


def get_params():
    p = Params()
    p.git_commit = False
    return p


def test_trainer():
    trainer = MyTrainer(get_params())

    trainer.params.eidx = 3
    fn = trainer.save_keypoint()

    trainer.params.eidx = 0
    trainer.load_checkpoint(fn)
    assert trainer.params.eidx == 3


def test_callbacks():
    trainer = MyTrainer(get_params())

    assert 1 in list(trainer._callback_set)[0].mark



from collections.abc import Iterable as CIterable
from itertools import chain
from typing import Callable, List, Union, Iterable, overload
from torch.utils.data import Dataset


class _Value:
    """
    base class for a sample wrap
    """

    def __init__(self, value, name=None):
        self.value = value
        self.name = name


class X(_Value): pass


class Y(_Value): pass


class ID(_Value): pass


class Delegate:
    """
    对数据形式的代理

    一般而言，只使用DatasetBuilder已经足够表示所有的数据形式，但凡事总有例外。例如，如果要构造一个Triplet（一般包含
    一个锚样本（anchor），一个正例（positive），一个负例（negtive）），那么正例和负例样本的index和Dataset中要取的index
    是不一样的，此时没有很好的方法来实现在DatasetBuilder中构建Triplet三元组——除非继承另外实现，但我不喜欢这种方式，
    于是就诞生了Delegate类。

    Examples:
    --------
    你唯一需要的是根据你的需求在__init__() 中初始化变量，并重写 __call__() 方法，在其中构建你的数据形式
    
    class TripletDelegate(Delegate):
        def __init__(self, xs, ys, cls_num=10) -> None:
            super().__init__()
            self.xs = xs
            self.ys = ys
            self.id_dict = defaultdict(list)
            for i, target in enumerate(ys):
                self.id_dict[target].append(i)
            self.neg_cls_ids = [[i for i in range(cls_num) if i != j] for j in range(cls_num)]

        def __call__(self, index, builder=None) -> Union[X, Y, ID, None, Iterable[Value]]:
            pos_cls = self.ys[index]
            neg_cls = random.choice(self.neg_cls_ids[pos_cls])
            pos_id, neg_id = random.choice(self.id_dict[pos_cls]), random.choice(self.id_dict[neg_cls])
            pos_img = self.xs[pos_id]
            neg_img = self.xs[neg_id]

        return self.X_(pos_img), self.X_(neg_img), self.Y_(neg_cls)

    Notes:
    -------
    注意，你返回的数据因为需要识别是input（X）还是label（Y），因此你需要在你返回的数据外包裹一层Value的子类X/Y/ID 来
    辅助识别；这些子类已经在Delegate内部写好了包裹方法，即
    >>> self.X_(...), self.X_(...), self.Y_(...)

    当然，你也可以手动引入这三个类：
    >>> from thexp.frame.builder import X,Y,ID

    """

    def X_(self, value, name=None):
        """
        用于包裹数据，辅助 DatasetBuilder 识别
        Args:
            value: 作为X的单个样本
            name: 当 DatasetBuilder 的输出形式为dict时，名字是必需的

        Returns:
            X(value, name)
        """
        return X(value, name)

    def Y_(self, value, name=None):
        """
        用于包裹数据，辅助 DatasetBuilder 识别
        Args:
            value: 作为Y的单个标签
            name: 当 DatasetBuilder 的输出形式为dict时，名字是必需的

        Returns:
            X(value, name)
        """
        return Y(value, name)

    def ID_(self, value, name=None):
        """
        用于包裹数据，辅助 DatasetBuilder 识别
        Args:
            value: 作为 index 的单个序号
            name: 当 DatasetBuilder 的输出形式为dict时，名字是必需的

        Returns:
            X(value, name)
        """
        return ID(value, name)

    def __len__(self):
        return 0

    def __call__(self, index, builder=None) -> Union[X, Y, ID, None, Iterable[_Value]]:
        raise NotImplementedError()

    def subset(self, indices):
        pass


class _placehold:
    """用于DatasetBuilder内部"""

    def __init__(self,
                 transform=None,
                 name=None, source=None, value=None):
        self.transform = transform
        self.name = name
        self.value = value
        self.source = source


class _delegate_placehold:
    """用于DatasetBuilder内部"""

    def __init__(self,
                 delegate: Delegate,
                 transform: Callable = None, target_transform: Callable = None,
                 name=None):
        self.delegate = delegate
        self.transform = transform
        self.target_transform = target_transform
        self.name = name


class DatasetBuilder(Dataset):
    """

    """

    def __init__(self, xs=None, ys=None, indices=None):
        self._dataset_len = None
        self._ipts = {}
        self._zip = False
        self._indices = indices
        self._re_indices = None
        self._placeholds = {'x': [], 'y': []}
        self._delegates = []  # type: List[_delegate_placehold]
        self._name = 'Dataset'
        self._source_names = set()
        self._data_names = set()
        self._delegate_names = set()
        self._add_id = False
        if xs is not None:
            self.add_inputs(xs)
        if ys is not None:
            self.add_labels(ys)

    def __str__(self):
        if self._zip:
            from thexp.base_classes.attr import attr
            res = attr()
            if self._add_id:
                res['ids'] = 'ids'
            for ph in chain(self._placeholds['x'], self._placeholds['y']):  # type:_placehold
                res[ph.name] = ph.source
            for dph in self._delegates:  # type: _delegate_placehold
                res[dph.name] = dph.delegate.__class__.__name__

            return str(res)
        else:
            ids = []
            xs = []
            ys = []
            deles = []
            if self._add_id:
                ids.append('ids')
            for ph in self._placeholds['x']:
                xs.append(ph.name)
            for ph in self._placeholds['y']:
                ys.append(ph.name)
            for dph in self._delegates:  # type: _delegate_placehold
                deles.append(dph.name)

            return str("{}: ({}, {}, {}, {})".format(
                self._name,
                ', '.join(ids),
                ', '.join(xs),
                ', '.join(ys),
                ', '.join(deles)))

    __repr__ = __str__

    def name(self, name: str):
        self._name = name
        return self

    def __len__(self) -> int:
        if self._re_indices is not None:
            return self.sample_num
        if self._indices is not None:
            return len(self._indices)
        return self._dataset_len

    def __getitem__(self, index: int):
        if self._re_indices is not None:
            if index >= self.real_num:
                index = index - self.real_num
                self._re_indices[index] = (self._re_indices[index] + self.vitu_num) % self.real_num
                index = self._re_indices[index]

        if self._indices is not None:
            index = self._indices[index]

        ids = []
        xs = []
        ys = []

        for ph in self._placeholds['x']:  # type:_placehold
            source = self._ipts[ph.source]
            x = source[index]
            if ph.transform is not None:
                x = ph.transform(x)
            ph.value = x
            xs.append(ph)
        for ph in self._placeholds['y']:
            source = self._ipts[ph.source]
            y = source[index]
            if ph.transform is not None:
                y = ph.transform(y)
            ph.value = y
            ys.append(ph)
        for dph in self._delegates:  # type:_delegate_placehold
            items = dph.delegate(index, self)
            if isinstance(items, _Value):
                items = [items]
            if isinstance(items, CIterable):
                for item in items:
                    assert isinstance(item, _Value), 'want instance of _Value, but {}'.format(item)
                    if self._zip:
                        assert item.name is not None, 'value must have name in zip mode'

                    if isinstance(item, X):
                        if dph.transform is not None:
                            item.value = dph.transform(item.value)
                        xs.append(_placehold(name=item.name, value=item.value))
                    if isinstance(item, Y):
                        if dph.target_transform is not None:
                            item.value = dph.target_transform(item.value)
                        ys.append(_placehold(name=item.name, value=item.value))
                    if isinstance(item, ID):
                        ids.append(_placehold(name=item.name, value=item.value))

        if self._zip:
            from thexp.base_classes.attr import attr
            res = attr()
            if self._add_id:
                res['_indexs'] = index
            for ph in ids:
                res[ph.name] = ph.value
            for ph in xs:
                res[ph.name] = ph.value
            for ph in ys:
                res[ph.name] = ph.value

            return res
        else:
            ids = [i.value for i in ids]
            xs = [i.value for i in xs]
            ys = [i.value for i in ys]
            if self._add_id:
                ids.insert(0, index)

            return [*ids, *xs, *ys]

    def _check_len(self, values):
        ilen = len(values)
        if self._dataset_len is None:
            self._dataset_len = ilen
        else:
            assert self._dataset_len == ilen, 'Current length of source is {}, but added length of {}'.format(
                self._dataset_len, ilen)

    def _check_source_name(self, name):
        assert name not in self._source_names, 'source name "{}" duplicates'.format(name)
        self._source_names.add(name)

    def _check_data_name(self, name):
        assert name not in self._data_names, 'data name "{}" duplicates'.format(name)
        self._data_names.add(name)

    def _check_delegate_name(self, name):
        assert name not in self._delegate_names, 'delegate name "{}" duplicates'.format(name)
        self._delegate_names.add(name)

    def add_inputs(self, inputs, source_name='xs'):
        self._check_source_name(source_name)
        self._check_len(inputs)
        self._ipts[source_name] = inputs
        return self

    def add_labels(self, labels, source_name='ys'):
        self._check_source_name(source_name)
        self._check_len(labels)
        self._ipts[source_name] = labels
        return self

    def toggle_id(self, toggle=None):
        if toggle is None:
            toggle = not self._add_id
        self._add_id = toggle
        return self

    def add_x(self, transform=None, name=None, source='xs'):
        if name is None:
            name = 'input_{}'.format(len(self._placeholds['x']))
        self._check_data_name(name)
        self._placeholds['x'].append(_placehold(transform=transform, name=name, source=source))
        return self

    def add_y(self, transform=None, name=None, source='ys'):
        if name is None:
            name = 'label_{}'.format(len(self._placeholds['y']))

        self._check_data_name(name)
        self._placeholds['y'].append(_placehold(transform=transform, name=name, source=source))
        return self

    def add_delegate(self, delegate: Delegate, transform=None, target_transform=None, name=None):
        # self.d
        if name is None:
            name = 'delegate_{}'.format(len(self._delegates))
        if self._indices is not None:
            delegate.subset(self._indices)

        self._check_delegate_name(name)
        self._check_len(delegate)
        self._delegates.append(_delegate_placehold(delegate, transform, target_transform, name))
        return self

    def zip_mode(self):
        self._zip = True
        return self

    def chain_mode(self):
        self.chain = True
        return self

    def subset(self, indices):
        self._indices = indices
        for delegate in self._delegates:
            delegate.delegate.subset(indices)
        return self

    def virtual_sample(self, sample_num: int):
        self.sample_num = sample_num
        if self._indices is None:
            self.real_num = self._dataset_len
        else:
            self.real_num = len(self._indices)
        self.vitu_num = self.sample_num - self.real_num
        self._re_indices = list(range(self.sample_num - self.real_num))
        return self

    @overload
    @staticmethod
    def DataLoader_from_dataset(dataset: Dataset, batch_size=1, shuffle=False, sampler=None,
                                batch_sampler=None, num_workers=0, collate_fn=None,
                                pin_memory=False, drop_last=False, timeout=0,
                                worker_init_fn=None, multiprocessing_context=None):
        pass

    @staticmethod
    def DataLoader_from_dataset(*args, **kwargs):
        from thexp.contrib.data import DataLoader
        return DataLoader(*args, **kwargs)

    @overload
    def DataLoader(self, batch_size=1, shuffle=False, sampler=None,
                   batch_sampler=None, num_workers=0, collate_fn=None,
                   pin_memory=False, drop_last=False, timeout=0,
                   worker_init_fn=None, multiprocessing_context=None):
        pass

    def DataLoader(self, *args, **kwargs):
        from thexp.contrib.data import DataLoader
        return DataLoader(self, *args, **kwargs)



"""

"""
import json
import os
import re
from collections import namedtuple
from typing import List

from ..utils.paths import listdir_by_time
import torch

ckpt_tuple = namedtuple("Checkpoint", ["checkpoint", 'info'])


class Saver:
    _ckpt_fn_templete = "{}{:06d}.ckpt"
    _model_fn_templete = "model.{}{:06d}.pth"
    re_fn = re.compile("^[0-9]{7}\.ckpt$")
    re_keep_fn = re.compile("^keep.[0-9]{7}\.ckpt$")

    def __init__(self, ckpt_dir, max_to_keep=3):
        self.info = {}
        self.ckpt_dir = ckpt_dir
        self.max_to_keep = max_to_keep
        os.makedirs(ckpt_dir, exist_ok=True)

    def _build_checkpoint_name(self, epoch=0, replacement: bool = False, lasting=False):
        i = 0

        def build_fn():
            fn = Saver._ckpt_fn_templete.format(i, epoch)
            if lasting:
                fn = "keep.{}".format(fn)
            return fn

        absfn = os.path.join(self.ckpt_dir, build_fn())
        while replacement == False and os.path.exists(absfn):
            i += 1
            if i >= 9:
                if lasting:
                    kfns = self.find_keypoints()
                else:
                    kfns = self.find_checkpoints()
                return kfns[-1]

            absfn = os.path.join(self.ckpt_dir, build_fn())
        return absfn

    def _build_model_name(self, epoch: int = 0) -> str:
        fn = os.path.join(self.ckpt_dir, Saver._model_fn_templete.format(0, epoch))
        return fn

    def _check_max_checkpoint(self):
        fs = self.find_checkpoints()
        while len(fs) > self.max_to_keep:
            self.check_remove(fs[-1])
            fs.pop()

    def find_checkpoints(self) -> List[str]:
        """
        find all checkpoint saved in the save dir.
        Returns:
        """

        fs = listdir_by_time(self.ckpt_dir)  # 按创建时间排序
        fs = [os.path.join(self.ckpt_dir, i) for i in fs if re.search(Saver.re_fn, i) is not None]
        return fs

    def find_models(self) -> List[str]:
        fs = listdir_by_time(self.ckpt_dir)  # 按创建时间排序
        fs = [i for i in fs if i.endswith(".pth")]
        return fs

    def find_keypoints(self) -> List[str]:
        """
        find all "keeped" checkpoint saved in the save dir.
        :return:
        """
        fs = listdir_by_time(self.ckpt_dir)  # 按创建时间排序
        fs = [os.path.join(self.ckpt_dir, i) for i in fs if re.search(Saver.re_keep_fn, i) is not None]
        return fs

    def save_model(self, epoch: int, state_dict, extra_info: dict = None) -> str:
        """
        命名格式为 model.{epoch}.pth , {epoch} 将padding到7位，默认为0
        重复保存会覆盖
        :param epoch:
        :param state_dict:
        :param extra_info:
        :return:
        """
        fn = self._build_model_name(epoch)
        json_fn = "{}.json".format(fn)
        torch.save(state_dict, fn)

        if extra_info is None:
            extra_info = dict()
        extra_info["fn"] = fn

        with open(json_fn, "w", encoding="utf-8") as w:
            json.dump(extra_info, w, indent=2)
        return fn

    def save_checkpoint(self, epoch, state_dict, extra_info: dict = None, replacement: bool = False,
                        lasting=False) -> str:
        """
        命名格式为 {epoch}.ckpt ，如果replacement=True，则命名格式为 keep.{epoch}.ckpt
        :param epoch:
        :param state_dict:
        :param extra_info: 额外信息，会以json格式保存在同模型名+ '.json'下
        :param replacement: 如果命名重复是否删除
        :param lasting: 是否受 max_to_keep 影响删除，默认为False
        :return:
        """
        fn = self._build_checkpoint_name(epoch, replacement, lasting)

        json_fn = "{}.json".format(fn)
        torch.save(state_dict, fn)
        if extra_info is None:
            extra_info = dict()
        extra_info["fn"] = fn

        with open(json_fn, "w", encoding="utf-8") as w:
            json.dump(extra_info, w, indent=2)

        if not lasting:
            self._check_max_checkpoint()

        return fn

    def save_keypoint(self, val, state_dict, extra_info: dict = None, replacement: bool = False) -> str:
        return self.save_checkpoint(val, state_dict, extra_info, replacement, True)

    def _check_isint_or_str(self, val) -> bool:
        assert type(val) in {int, str}
        if isinstance(val, int):
            return True
        return False

    def load_latest_checkpoint(self, dir=None):
        if dir is None:
            dir = self.ckpt_dir

        fs = listdir_by_time(dir)
        fs = [os.path.join(dir, i) for i in fs if i.endswith(".ckpt")]
        if len(fs) == 0:
            return ckpt_tuple(None, None)
        return self.load_state_dict(fs[-1])

    def load_state_dict(self, fn: str) -> ckpt_tuple:
        """

        Args:
            fn: fn: rel path or abs path

        Returns:
            A namedtuple("Checkpoint", ["checkpoint", 'info']) instance, include two items, first item is a state_dict,
            second item is information attached when stored the state_dict
        """
        path = self._guess_abs_path(fn)

        if os.path.exists(path):
            ckpt = torch.load(fn)
        else:
            ckpt = None
        info_path = "{}.json".format(path)
        if os.path.exists(info_path):
            with open(info_path, 'r', encoding='utf-8') as r:
                info = json.load(r)
        else:
            info = None

        return ckpt_tuple(ckpt, info)

    def _guess_abs_path(self, fn: str) -> str:
        if os.path.basename(fn) == fn:
            path = os.path.join(self.ckpt_dir, fn)
        else:
            path = fn
        return path

    def check_remove(self, fn: str, with_json=True):
        if os.path.exists(fn):
            os.remove(fn)
        if with_json:
            jfn = "{}.json".format(fn)
            if os.path.exists(jfn):
                os.remove(jfn)

    def clear_models(self):
        fns = filter(lambda x: x.endswith(".pth"),
                     listdir_by_time(self.ckpt_dir))
        for i in fns:
            self.check_remove(os.path.join(self.ckpt_dir, i))

    def clear_checkpoints(self):
        for i in self.find_checkpoints():
            self.check_remove(i)

    def clear_keypoints(self):
        fns = filter(lambda x: x.endswith(".ckpt") and x.startswith("keep."),
                     listdir_by_time(self.ckpt_dir))
        for i in fns:
            self.check_remove(os.path.join(self.ckpt_dir, i))

    def summary(self):
        print("checkpoints:")
        print(" || ".join(self.find_checkpoints()))
        print("keppoints:")
        print(" || ".join(self.find_keypoints()))
        print("models:")
        print(" || ".join(self.find_models()))



"""
Methods about git.
"""
import json
import os
import sys
from functools import lru_cache
from typing import List
from uuid import uuid4

from git import Git, Repo
from thexp import __VERSION__
from thexp.utils.paths import renormpath
from .dates import curent_date
from ..globals import _GITKEY, _OS_ENV, _FNAME

torch_file = ['*.pth', '*.npy', '*.ckpt']
thexp_gitignores = ['.thexp/', _FNAME.repo, _FNAME.expsdirs, '.idea/', '*.thexp.*'] + torch_file

py_gitignore = "\n".join(['# Byte-compiled / optimized / DLL files', '__pycache__/', '*.py[cod]',
                          '*$py.class', '', '# C extensions', '*.so', '', '# Distribution / packaging',
                          '.Python', 'build/', 'develop-eggs/', 'dist/', 'downloads/', 'eggs/', '.eggs/',
                          'lib/', 'lib64/', 'parts/', 'sdist/', 'var/', 'wheels/', 'pip-wheel-metadata/',
                          'share/python-wheels/', '*.egg-info/', '.installed.cfg', '*.egg', 'MANIFEST',
                          '', '# PyInstaller',
                          '#  Usually these files are written by a python script from a template',
                          '#  before PyInstaller builds the exe, so as to inject date/other infos into '
                          'it.',
                          '*.manifest', '*.spec', '', '# Installer logs', 'pip-log.txt',
                          'pip-delete-this-directory.txt', '', '# Unit test / coverage reports',
                          'htmlcov/', '.tox/', '.nox/', '.coverage', '.coverage.*', '.cache',
                          'nosetests.xml', 'coverage.xml', '*.cover', '.hypothesis/', '.pytest_cache/',
                          '', '# Translations', '*.mo', '*.pot', '', '# Django stuff:', '*.log',
                          'local_settings.py', 'db.sqlite3', '', '# Flask stuff:', 'instance/',
                          '.webassets-cache', '', '# Scrapy stuff:', '.scrapy', '',
                          '# Sphinx documentation', 'docs/_build/', '', '# PyBuilder', 'target/', '',
                          '# Jupyter Notebook', '.ipynb_checkpoints', '', '# IPython',
                          'profile_default/', 'ipython_config.py', '', '# pyenv', '.python-version', '',
                          '# celery beat schedule file', 'celerybeat-schedule', '',
                          '# SageMath parsed files', '*.sage.py', '', '# Environments', '.env', '.venv',
                          'env/', 'venv/', 'ENV/', 'env.bak/', 'venv.bak/', '',
                          '# Spyder project settings', '.spyderproject', '.spyproject', '',
                          '# Rope project settings', '.ropeproject', '', '# mkdocs documentation',
                          '/site', '', '# mypy', '.mypy_cache/', '.dmypy.json', 'dmypy.json', '',
                          '# Pyre type checker', '.pyre/'] + thexp_gitignores)


def _set_default_config(repo: Repo, names: List[str]):
    """
    set default git config, including:
    1. [expsdir], represent the root path of the experiment extra data(saved by thexp.frame.Experiment)
    of this repository

    2. [uuid], to avoid duplicate name, all repository will have another 2-bit hash value produced by uuid.

    3. [projname], record the directory name of this repository to guarantee the key flag of this repository won't change.
    """
    if len(names) == 0:
        return {}

    def _expsdir(repo):
        from .paths import global_config
        config = global_config()
        expsdir = config.get(_GITKEY.expsdir, renormpath(os.path.join(repo.working_dir, '.thexp/experiments')))
        return expsdir

    _default = {
        _GITKEY.uuid: lambda *args: uuid4().hex[:2],
        _GITKEY.expsdir: lambda repo: _expsdir(repo),
        _GITKEY.projname: lambda repo: renormpath(os.path.basename(repo.working_dir))
    }

    writer = repo.config_writer()
    res = {}
    for name in names:
        value = _default[name](repo)
        writer.add_value(_GITKEY.section_name, name, value)
        res[name] = value

    writer.write()
    writer.release()

    return res


def _check_section(repo: Repo):
    """
    check if git config section([thexp]) exists.
    """
    writer = repo.config_writer()
    if not writer.has_section(_GITKEY.section_name):
        writer.add_section(_GITKEY.section_name)
    writer.write()
    writer.release()


@lru_cache()
def git_config(repo: Repo):
    """
    load the git config of the given repository.

    Because of functools.lru_cache, this time-cosuming function will only read git config once from disk
    for each repository. And it will then load from lru_cache.
    """
    reader = repo.config_reader()
    if not reader.has_section(_GITKEY.section_name):
        _check_section(repo)
        reader = repo.config_reader()
    reader.read()
    try:
        config = {k: v for k, v in reader.items(_GITKEY.section_name)}
    except:
        config = {}

    lack_names = [i for i in {_GITKEY.expsdir, _GITKEY.uuid, _GITKEY.projname} if i not in config]
    _updates = _set_default_config(repo, lack_names)

    config.update(_updates)

    return config


def check_gitignore(repo: Repo, force=False):
    """
    check if file `.gitignore`  have the needed ignored items for thexp.
    """
    rp = os.path.join(repo.working_dir, _FNAME.expsdirs)

    version_mark = os.path.join(repo.working_dir, _FNAME.gitignore_version)
    if os.path.exists(rp) and os.path.exists(version_mark) and not force:
        return

    old_marks = [f for f in os.listdir(repo.working_dir) if f.startswith('.thexp.')]
    for old_mark in old_marks:
        mark_fn = os.path.join(repo.working_dir, old_mark)
        if os.path.isfile(mark_fn):
            os.remove(mark_fn)

    with open(version_mark, 'w') as w:
        pass

    ignorefn = os.path.join(repo.working_dir, _FNAME.gitignore)
    if not os.path.exists(ignorefn):
        with open(ignorefn, 'w', encoding='utf-8') as w:
            w.write(py_gitignore)
        return True
    else:
        amend = False
        with open(ignorefn, 'r', encoding='utf-8') as r:
            lines = [i.strip() for i in r.readlines()]
            for item in thexp_gitignores:
                if item not in lines:
                    lines.append(item)
                    amend = True
        if amend:
            with open(ignorefn, 'w', encoding='utf-8') as w:
                w.write('\n'.join(lines))

    return amend


def git_config_syntax(value: str) -> str:
    """
    syntax value(especially file path) which will be stored in git config
    """
    return value.replace('\\\\', '/').replace('\\', '/')


@lru_cache()
def git_root(dir="./", ignore_info=False):
    """
    判断某目录是否在git repo 目录内（包括子目录），如果是，返回该 repo 的根目录
    :param dir:  要判断的目录。默认为程序运行目录
    :return: 如果是，返回该repo的根目录（包含 .git/ 的目录）
        否则，返回空
    """
    cur = os.getcwd()
    os.chdir(dir)
    try:
        res = Git().execute(['git', 'rev-parse', '--git-dir'])
    except Exception as e:
        if not ignore_info:
            print(e)
        res = None
    os.chdir(cur)
    return res


class branch:
    """
    用于配合上下文管理切换 git branch

    with branch(repo, branch):
        repo.index.commit('...')
    """

    def __init__(self, repo: Repo, branch: str):
        self.repo = repo
        self.old_branch = self.repo.head.reference
        self.branch = branch

    def __enter__(self):
        if self.branch not in self.repo.heads:
            head = self.repo.create_head(self.branch)
        else:
            head = self.repo.heads[self.branch]

        self.repo.head.reference = head
        # self.repo.head.reset(index=True, working_tree=True)
        return head

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.repo.head.reference = self.old_branch


def init_repo(dir='./'):
    """
    initialize a directory, including git init, thexp config and a initial commit.
    """
    path = git_root(dir, ignore_info=True)
    if path is not None:
        res = Repo(path)
    else:
        res = Repo.init(path)
    check_gitignore(repo=res, force=True)
    git_config(res)
    res.git.add('.')
    res.index.commit('initial commit')
    return res


@lru_cache()
def load_repo(dir='./') -> Repo:
    """
    Try to load git repository object of a directory.
    Args:
        dir: str, a directory path, default is the current working dir.
        if dir is a repository dir, then a git.Repo object will be retured.
        if not, some you can type a path to init it, or type '!' to cancel init it.

    Returns:
        git.Repo object or None if dir not have git repository and cancel to init it.
    """

    path = git_root(dir)

    if path is None:
        if _OS_ENV.IGNORE_REPO not in os.environ:
            print("fatal: not a git repository (or any of the parent directories)")
            print("-----------------------")
            path = input("type root path to init this project, \n(default: {}, type '!' to ignore".format(os.getcwd()))
        else:
            print("Variable 'repo' will be a None object. "
                  "Any operation that need this repo may cause Exception.")
            return None

        if '!' in path or '！' in path:
            print("Variable 'repo' will be a None object. "
                  "Any operation that need this repo may cause Exception.")
            return None

        res = Repo.init(path)
        check_gitignore(repo=res, force=True)
        # check_gitconfig(repo=res, force=True)
        res.git.add('.')
        res.index.commit('initial commit')
    else:
        res = Repo(path)
        amend = check_gitignore(repo=res, force=False)
        if amend:
            res.git.add(_FNAME.gitignore)
            res.index.commit('fix gitignore')
    return res


_commits_map = {}


def commit(repo: Repo, key=None, branch_name=_GITKEY.thexp_branch):
    """"""
    if key is not None and key in _commits_map:
        return _commits_map[key]

    with branch(repo, branch_name):
        repo.git.add(all=True)
        commit_date = curent_date()
        commit_info = dict(
            date=commit_date,
            args=sys.argv,
            environ="jupyter" if "jupyter_core" in sys.modules else "python",
            version=sys.version,
        )
        commit_ = repo.index.commit(json.dumps(commit_info, indent=2))
    if key is not None:
        _commits_map[key] = commit_
    return commit_



from collections import defaultdict

import numpy as np
import torch


def create_id_dict(targets):
    if isinstance(targets, torch.Tensor):
        targets = targets.detach().cpu().numpy()
    else:
        targets = np.array(targets)

    id_dict = defaultdict(list)
    for i, target in enumerate(targets):
        id_dict[target].append(i)

    return id_dict





"""
thexp provides an easy git-like config system.

globals : can be used in all repository.
repository : only used in each repository.
running : only used in current run-time.

A global instance is exposed, here is some examples:

```python
from thexp import globs
globs.add_value('dataset','path/to/dataset/root/',globs.LEVEL.globals)
root = globs['dataset']
```

you can get all config by use `globs.items()`
"""
from pprint import pformat
from typing import Any

from thexp.utils.paths import global_config, write_global_config
from ..globals import _CONFIGL, _GITKEY


class Globals:
    LEVEL = _CONFIGL

    def __init__(self):
        self._configs = [
            Config(_CONFIGL.running),
            Config(_CONFIGL.repository),
            Config(_CONFIGL.globals),
        ]

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("_"):
            super().__setattr__(name, value)
        else:
            self.__setitem__(name, value)

    def __getattr__(self, item):
        return self.__getitem__(item)

    def __getitem__(self, item):
        for config in self._configs:
            if item in config:
                return config[item]

    def __setitem__(self, key, value):
        self._configs[0][key] = value

    def add_value(self, key, value, level=_CONFIGL.globals):
        if level == _CONFIGL.globals:
            self._configs[2][key] = value
        elif level == _CONFIGL.repository:
            self._configs[1][key] = value
        elif level == _CONFIGL.running:
            self._configs[0][key] = value
        else:
            assert False, 'level name error {}'.format(level)

    def get_value(self, key, level=_CONFIGL.globals, default=None):
        try:
            if level == _CONFIGL.globals:
                return self._configs[2][key]
            elif level == _CONFIGL.repository:
                return self._configs[1][key]
            elif level == _CONFIGL.running:
                return self._configs[0][key]
        except:
            return default
        assert False, 'level name error {}'.format(level)

    def items(self):
        return {
            _CONFIGL.globals: self._configs[2].items(),
            _CONFIGL.repository: self._configs[1].items(),
            _CONFIGL.running: self._configs[0].items(),
        }

    def __repr__(self):
        return "Globals({})".format(pformat({
            _CONFIGL.globals: self._configs[2].items(),
            _CONFIGL.repository: self._configs[1].items(),
            _CONFIGL.running: self._configs[0].items(),
        }))

    @property
    def runtime_config(self):
        return self._configs[0]

    @property
    def repository_config(self):
        return self._configs[1]

    @property
    def globals_config(self):
        return self._configs[2]

    @property
    def repo_root(self):
        from ..utils.repository import git_root
        return git_root()


class Config:
    """
    试验配置，根据等级分为用户级（整个用户），repo级（当前项目），实验级（当次运行）
    """
    config_levels = {_CONFIGL.running, _CONFIGL.repository, _CONFIGL.globals}

    def __init__(self, config_level):
        assert config_level in Config.config_levels, 'config level must in {}'.format(Config.config_levels)
        self._config_level = config_level
        if config_level == _CONFIGL.running:
            self._config_dict = {}
        elif config_level == _CONFIGL.repository:
            self._repo = None
            self._config_dict = None
        elif config_level == _CONFIGL.globals:
            self._config_dict = global_config()

    @property
    def repo(self):
        if self.config_level == _CONFIGL.repository:
            from ..utils.repository import load_repo
            self._repo = load_repo()
        return self._repo

    @property
    def repo_config(self):
        if not self.repo_level or self.repo is None:
            return {}
        if self._config_dict is None:
            from ..utils.repository import git_config
            self._config_dict = git_config(self.repo)
        return self._config_dict

    @property
    def config_level(self):
        return self._config_level

    @property
    def running_level(self):
        return self._config_level == _CONFIGL.running

    @property
    def globals_level(self):
        return self._config_level == _CONFIGL.globals

    @property
    def repo_level(self):
        return self._config_level == _CONFIGL.repository

    def __setitem__(self, key, value: str):
        """
        key 和 value尽量简洁
        :param key:
        :param value:
        :return:
        """
        key = str(key)
        if self._config_level == _CONFIGL.globals:
            self._config_dict[key] = value
            write_global_config(self._config_dict)
        elif self._config_level == _CONFIGL.repository:
            from thexp.utils.repository import git_config_syntax
            value = git_config_syntax(value)
            repo = self.repo
            if repo is not None:
                writer = repo.config_writer()
                writer.add_value(_GITKEY.section_name,key,value)
                writer.write()
                writer.release()
            self._config_dict[key] = value
        elif self._config_level == _CONFIGL.running:
            self._config_dict[key] = value

    def __getitem__(self, key):
        key = str(key)
        if self._config_level in {_CONFIGL.globals, _CONFIGL.running}:
            if key not in self._config_dict:
                raise AttributeError(key)
            return self._config_dict[key]
        elif self._config_level == _CONFIGL.repository:
            if key not in self.repo_config:
                raise AttributeError(key)
            return self.repo_config[key]

    def items(self):
        if self._config_level == _CONFIGL.running:
            return self._config_dict.items()
        elif self._config_level == _CONFIGL.repository:
            return self.repo_config.items()
        elif self._config_level == _CONFIGL.globals:
            return self._config_dict.items()

    def __contains__(self, item):
        try:
            _ = self[item]
            return True
        except:
            return False

    def __repr__(self) -> str:
        return pformat(self.items())


globs = Globals()



from typing import Union, Dict, List, Callable


def regist_func(val: Union[Dict[str, Callable], List[Callable]], name_=None):
    def wrap(func):
        if name_ is None:
            name = func.__name__
        else:
            name = name_
        if isinstance(val, dict):
            val[name] = func
        elif isinstance(val, list):
            val.append(func)

        return func

    return wrap



"""

"""

from torch.utils.data.dataloader import DataLoader as _DataLoader

# from thexp.utils.decorators.deprecated import deprecated
from thexp import __VERSION__


class DataLoader(_DataLoader):
    """
    用于可调整 batch_size 的DataLoader，其需求基于

    Samuel L. Smith, Pieter-Jan Kindermans, Chris Ying, Quoc V. Le, Don't Decay the Learning Rate, Increase the Batch Size
    https://arxiv.org/abs/1711.00489
    """

    def set_batch_size(self, batch_size):
        self.batch_sampler.batch_size = batch_size

    def DataBundler(self, add=True, name=None):
        from thexp import DataBundler
        bundler = DataBundler()
        if add:
            return bundler.add(self, name)
        else:
            return bundler.cycle(self, name)



"""

"""
import copy
import json
import os
import pprint as pp
import warnings
from collections import defaultdict
from collections.abc import Iterable
from datetime import timedelta
from typing import Any, overload

import fire
import torch

from ..base_classes.attr import attr
from ..base_classes.defaults import default
from ..base_classes.errors import BoundCheckError, NewParamWarning
from ..base_classes.params_vars import ParamsFactory, OptimParams, OptimMixin
from ..utils.environ import ENVIRON_
from ..calculate import schedule
from ..decorators.deprecated import deprecated


class BaseParams(OptimMixin):
    ENV = ENVIRON_

    class SCHE:
        Cos = schedule.CosSchedule
        Linear = schedule.LinearSchedule
        Log = schedule.LogSchedule
        Exp = schedule.ExpSchedule
        Power = schedule.PowerDecaySchedule
        Const = schedule.ConstantSchedule

        PeriodCos = schedule.PeriodCosSchedule
        PeriodHalfCos = schedule.PeriodHalfCosSchedule
        PeriodLinear = schedule.PeriodLinear
        PeriodTriangle = schedule.PeriodTriangleSchedule

        List = schedule.ScheduleList

    def __init__(self):
        self._param_dict = attr()
        self._repeat = None
        self._bound = {}
        self._lock = False
        # self._bind = defaultdict(list)

    def __getitem__(self, item):
        return self.__getattr__(item)

    def __setitem__(self, key, value):
        key = str(key)
        self.__setattr__(key, value)

    def __contains__(self, item):
        return item in self._param_dict

    def __getstate__(self):
        return {
            '_param_dict': self._param_dict,
            '_repeat': self._repeat,
            # '_bound': self._bound,
            '_lock': self._lock,
            # '_bind': self._bind,
        }

    def __setstate__(self, d):
        self._param_dict = d['_param_dict']
        self._repeat = d['_repeat']
        # self._bound = d['_bound']
        self._lock = d['_lock']
        # self._bind = d['_bind']

    def __setattr__(self, name: str, value: Any) -> None:
        """
        1. check constrain
        2. check if is default and not exists
        3. check bind
        """
        from ..base_classes.defaults import default
        if name.startswith("_"):
            super().__setattr__(name, value)
        else:
            res = self._check(name, value)
            if res is not None and not res:
                raise BoundCheckError("param '{}' checked failed.".format(name))

            if isinstance(value, default):  # 设置默认值
                # only set default value when name not exists
                if name not in self._param_dict:
                    if value.warn:
                        warnings.warn(
                            "'{}' is a new param,please check your spelling. It's more recommended to define in advance.".format(
                                name))
                    value = value.default
                    self._param_dict[name] = value
            else:
                self._param_dict[name] = value
            # if name in self._bind:
            # for _v, _bind_k, _bind_v in self._bind[name]:
            #     if callable(_v):
            #         _bind_v = _v(value)
            #         self.__setattr__(_bind_k, _bind_v)
            #     elif _v == value:
            #         self.__setattr__(_bind_k, _bind_v)

    def __getattr__(self, item):
        return self._param_dict.__getattr__(item)

    def __repr__(self):
        return "{}".format(self.__class__.__name__) + pp.pformat([(k, v) for k, v in self._param_dict.items()])

    __str__ = __repr__

    def __delattr__(self, name: str) -> None:
        if name.startswith("_"):
            super().__delattr__(name)
        else:
            self._param_dict.pop(name)

    def __delitem__(self, key):
        key = str(key)
        self.__delattr__(key)

    def __eq__(self, other):
        if isinstance(other, BaseParams):
            return self.hash() == other.hash()
        return False

    def _check(self, name, value):
        if isinstance(value, default):
            value = value.default
        if name in self._bound:
            self._bound[name](value)

    def initial(self):
        pass

    def arange(self, k, default, left=float("-inf"), right=float("inf")):
        """
        constrain value within a continuous interval.
        Args:
            k: key of the value
            default: default value
            left: left interval
            right: right interval

        Returns:
            default value

        Notes:
        ---------
        you can directly call this function to create the key with the constrained value, but it's better to
        receive this value so that IDE can recognizes it and provides auto-complete.
        """

        def check(x):
            if x < left or x > right:
                raise BoundCheckError("param '{}' must be within range({}, {})".format(k, left, right))

        self._bound[k] = check
        self[k] = default
        return default

    def choice(self, k, *choices):
        """
        constrain value within a discrete set.
        Args:
            k: key of the value
            *choices: value can be used for key

        Returns:
            the first value of choices

        """

        def check(x):
            if x not in choices:
                raise BoundCheckError("param '{}' is enum of {}".format(k, choices))

        self._bound[k] = check
        self[k] = choices[0]
        return choices[0]

    def lambda_bound(self, k, default, check_lmd):
        """"""
        pass
        # self._bound[k] = check_lmd
        # self[k] = default
        # return default

    @deprecated('1.5.1', '1.6', 'Call initial() to create dynamic parameter is a better choice.')
    def bind(self, k, v, bind_k, bind_v):
        """
        Link the change in the value of one key with another.
        When params[k] is set to v, params[bind_k] will be set to bind_v automatic
        """
        # self._bind[k].append((v, bind_k, bind_v))
        pass

    @deprecated('1.5.1', '1.6', 'Call initial() to create dynamic parameter is a better choice.')
    def dynamic_bind(self, k, bind_k, dynamic_func):
        # self._bind[k].append((dynamic_func, bind_k, None))
        pass

    def grid_search(self, key, iterable: Iterable):
        """"""
        for v in iterable:
            res = self._copy()
            res[key] = v
            yield res

    def _copy(self):
        res = BaseParams()
        res._param_dict = copy.copy(self._param_dict)
        res._repeat = copy.copy(self._repeat)
        # res._bound = copy.copy(self._bound)
        res._lock = copy.copy(self._lock)
        # res._bind = copy.copy(self._bind)
        return res

    def grid_range(self, count):
        for i in range(count):
            res = self._copy()
            res._repeat = i
            yield res

    def from_args(self):
        """
        从命令行参数中设置参数值

        可选参数：
            _json, 接收一个json来作为该params的参数
            _help, 输出该Params已设置的参数

        Returns:

        """

        def func(**kwargs):
            if '_help' in kwargs:
                print(self)
                return

            if '_json' in kwargs:
                self.from_json(kwargs['_json'])
                return

            for k, v in kwargs.items():
                try:
                    self[k]
                except:
                    warnings.simplefilter('always', NewParamWarning)
                    warnings.warn(
                        "'{}' is a new param,please check your spelling.\n it's more recommended to define in advance.".format(
                            k))
                self[k] = v

        fire.Fire(func)
        return self

    def from_json(self, fn):
        """
        从 json 中获取参数值
        Args:
            fn:

        Returns:

        """
        if os.path.exists(fn):
            with open(fn, encoding='utf-8') as r:
                res = json.load(r)
                for k, v in res.items():
                    self[k] = v
        return self

    def to_json(self, fn: str):
        """
        以json格式保存文件，注意，该方法保存的json内容基于 attr 的jsonify() 方法，不可序列化的格式无法被保存
        Args:
            fn:

        Returns:

        """
        with open(fn, 'w', encoding='utf-8') as w:
            json.dump(self.inner_dict.jsonify(), w, indent=2)

    def items(self):
        return self._param_dict.items()

    def keys(self):
        for k in self._param_dict:
            yield k

    def update(self, dic: dict):
        """

        Args:
            dic:

        Returns:

        """
        for k, v in dic.items():
            self._param_dict[k] = v

        return self

    def hash(self) -> str:
        """
        返回对参数的定义顺序及其相应值的一个hash，理论上，两个Param 对象的hash方法返回的参数相同，
        则两者具有相同的参数和参数值

        Returns:

        """
        return self._param_dict.hash()

    def lock(self):
        """
        锁定当前配置，如果当前配置未 lock，那么当尝试获取未分配的参数时候，会返回一个空的字典
        如果锁定，则在尝试获取未分配参数时，会抛出 AttributeError(key)
        Returns:

        """
        self._lock = True

    @property
    def inner_dict(self) -> attr:
        return self._param_dict

    def get(self, k, default=None):
        """
        获取某值，如果不存在，则返回默认值
        Args:
            k:
            default:

        Returns:

        """
        if k in self:
            return self[k]
        else:
            return default

    @staticmethod
    def default(value: Any = None, warn=False):
        """
        默认值，分配值时，仅当当前key没有分配时，分配该值作为键值。否则，该值会被忽略

        Examples:
        >>> params.margin = params.default(0.5,True)
        >>> params.margin = params.default(0.3,True)
        >>> print(params.margin)

        Args:
            value: 要设置的值
            warn: 当设置默认值时，抛出警告

        Returns:
            default(value, warn)
        """
        from ..base_classes.defaults import default
        return default(value, warn)

    def create_schedule(self, schedule_type, start, end, **kwargs):
        return ParamsFactory

    def replace(self, **kwargs):
        self.update(kwargs)
        return self

    def contains(self, key: str):
        return key in self


class DistributionParams(BaseParams):
    def __init__(self):
        super().__init__()
        self.backend = 'nccl'
        self.distributed = False
        self.world_size = -1
        self.local_rank = -1  # if not -1, means will use
        self.init_method = 'env://'


class Params(BaseParams):
    Attr = attr

    def __init__(self):
        super().__init__()
        self.epoch = 10
        self.eidx = 1
        self.idx = 0
        self.global_step = 0
        self.device = "cuda:0" if torch.cuda.is_available() else "cpu"
        self.device_ids = []
        self.dataset = None
        self.architecture = None
        self.optim = None  # type:OptimParams
        self.git_commit = True
        self.tmp_dir = None  # type:str # set TMPDIR environment

        self.distributed = False
        self.world_size = -1
        self.local_rank = -1  # if not -1, means will use
        self.init_method = 'env://'

    def enable_distribution(self):
        pass

    @overload
    def init_process_group(self, backend,
                           init_method=None,
                           timeout=timedelta(minutes=30),
                           world_size=-1,
                           rank=-1,
                           store=None,
                           group_name=''):
        pass

    def init_process_group(self, *args, **kwargs):
        self.init_process_group_args = (args, kwargs)
        return self.init_process_group_args



"""

"""
__VERSION__ = "1.5.0.33"

from .frame import (
    Logger,
    Meter,
    AvgMeter,
    Params, BaseParams,
    Saver,
    RndManager,
    Delegate,
    DatasetBuilder,
    DataBundler,
    Trainer,
    callbacks,
    Experiment,
    globs)

from .analyse import Q, C
from .utils.environ import ENVIRON_ as ENV

import thexp.calculate  # initialize schedule attr classes
from .utils.memory import memory


"""

"""


class tree(dict):
    """Implementation of perl's autovivification feature."""

    def __getitem__(self, item):
        try:
            return dict.__getitem__(self, item)
        except KeyError:
            value = self[item] = type(self)()
            return value



import sys

if sys.platform == "win32" and sys.stdin.isatty():
    FOREGROUND_BLACK = 0x00  # black.
    FOREGROUND_DARKBLUE = 0x01  # dark blue.
    FOREGROUND_DARKGREEN = 0x02  # dark green.
    FOREGROUND_DARKSKYBLUE = 0x03  # dark skyblue.
    FOREGROUND_DARKRED = 0x04  # dark red.
    FOREGROUND_DARKPINK = 0x05  # dark pink.
    FOREGROUND_DARKYELLOW = 0x06  # dark yellow.
    FOREGROUND_DARKWHITE = 0x07  # dark white.
    FOREGROUND_DARKGRAY = 0x08  # dark gray.
    FOREGROUND_BLUE = 0x09  # blue.
    FOREGROUND_GREEN = 0x0a  # green.
    FOREGROUND_SKYBLUE = 0x0b  # skyblue.
    FOREGROUND_RED = 0x0c  # red.
    FOREGROUND_PINK = 0x0d  # pink.
    FOREGROUND_YELLOW = 0x0e  # yellow.
    FOREGROUND_WHITE = 0x0f  # white.
    FOREGROUND_PURPLE = FOREGROUND_PINK
    FOREGROUND_CYAN = FOREGROUND_PINK

    BACKGROUND_BLACK = 0x00
    BACKGROUND_DARKBLUE = 0x10  # dark blue.
    BACKGROUND_DARKGREEN = 0x20  # dark green.
    BACKGROUND_DARKSKYBLUE = 0x30  # dark skyblue.
    BACKGROUND_DARKRED = 0x40  # dark red.
    BACKGROUND_DARKPINK = 0x50  # dark pink.
    BACKGROUND_DARKYELLOW = 0x60  # dark yellow.
    BACKGROUND_DARKWHITE = 0x70  # dark white.
    BACKGROUND_DARKGRAY = 0x80  # dark gray.
    BACKGROUND_BLUE = 0x90  # blue.
    BACKGROUND_GREEN = 0xa0  # green.
    BACKGROUND_SKYBLUE = 0xb0  # skyblue.
    BACKGROUND_RED = 0xc0  # red.
    BACKGROUND_PINK = 0xd0  # pink.
    BACKGROUND_YELLOW = 0xe0  # yellow.
    BACKGROUND_WHITE = 0xf0  # white.
    BACKGROUND_PURPLE = BACKGROUND_PINK
    BACKGROUND_CYAN = BACKGROUND_PINK
else:
    FOREGROUND_BLACK = 30
    FOREGROUND_RED = 31
    FOREGROUND_GREEN = 32
    FOREGROUND_YELLOW = 33
    FOREGROUND_BLUE = 34
    FOREGROUND_PURPLE = 35
    FOREGROUND_CYAN = 36
    FOREGROUND_WHITE = 37
    FOREGROUND_SKYBLUE = FOREGROUND_BLUE
    FOREGROUND_PINK = FOREGROUND_PURPLE
    FOREGROUND_DARKBLUE = FOREGROUND_BLUE
    FOREGROUND_DARKGREEN = FOREGROUND_GREEN
    FOREGROUND_DARKSKYBLUE = FOREGROUND_BLUE
    FOREGROUND_DARKRED = FOREGROUND_RED
    FOREGROUND_DARKPINK = FOREGROUND_CYAN
    FOREGROUND_DARKYELLOW = FOREGROUND_YELLOW
    FOREGROUND_DARKWHITE = FOREGROUND_WHITE
    FOREGROUND_DARKGRAY = FOREGROUND_BLACK

    BACKGROUND_BLACK = 40
    BACKGROUND_RED = 41
    BACKGROUND_GREEN = 42
    BACKGROUND_YELLOW = 43
    BACKGROUND_BLUE = 44
    BACKGROUND_PURPLE = 45
    BACKGROUND_CYAN = 46
    BACKGROUND_WHITE = 47
    BACKGROUND_SKYBLUE = BACKGROUND_BLUE
    BACKGROUND_PINK = BACKGROUND_PURPLE
    BACKGROUND_DARKBLUE = BACKGROUND_BLUE
    BACKGROUND_DARKGREEN = BACKGROUND_GREEN
    BACKGROUND_DARKSKYBLUE = BACKGROUND_BLUE
    BACKGROUND_DARKRED = BACKGROUND_RED
    BACKGROUND_DARKPINK = BACKGROUND_PURPLE
    BACKGROUND_DARKYELLOW = BACKGROUND_YELLOW
    BACKGROUND_DARKWHITE = BACKGROUND_WHITE
    BACKGROUND_DARKGRAY = BACKGROUND_BLACK
    
MODE_NORMAL = 0
MODE_BOLD = 1
MODE_UNDERLINE = 4
MODE_BLINK = 5
MODE_INVERT = 7
MODE_HIDE = 8



"""
对各不同级别的实验日志存储信息的查看
"""
import json
import os
import pprint
import shutil
from functools import lru_cache
from typing import Dict

from thexp import Experiment
from thexp.base_classes.attr import attr
from thexp.globals import _INFOJ, _FNAME, _BUILTIN_PLUGIN, _GITKEY, _TEST_BUILTIN_STATE, _PLUGIN_DIRNAME, _PLUGIN_KEY
from .reader import BoardReader
from ..utils.dates import date_from_str


class _Viewer():
    def delete(self):

        dir_ = self.root
        if dir_ is None:
            return
        shutil.rmtree(dir_)

        from itertools import chain
        for attr_name in chain(dir(self.__class__), dir(self)):
            if attr_name.startswith('__'):
                continue
            try:
                attr = getattr(self, attr_name)
            except:
                attr = None
            try:
                if callable(attr):
                    setattr(self, attr_name, lambda *args, **kwargs: 'NoneType')
                else:
                    setattr(self, attr_name, None)
            except:
                pass
        self.__str__ = lambda: ''
        self.__repr__ = lambda: ''

    @property
    def root(self):
        raise NotImplementedError()

    def tree(self, depth: int = 0):
        raise NotImplementedError()

    def summary(self):
        return pprint.pprint(self.tree(0))


class ProjViewer(_Viewer):
    """
    以项目为单位的实验浏览类，用于查看所有 未删除 的 exp 对应的方法
    """

    def __init__(self, proj_dir: str):
        self._proj_dir = proj_dir
        self._repo = None
        self._repo_info = None

    def __iter__(self):
        for exp_viewer in self.exp_viewers:
            for test_viewer in exp_viewer:
                yield test_viewer

    def __repr__(self):
        return "ProjViewer(exps=[{}])".format(self.exp_names)

    @property
    def name(self):
        return os.path.basename(self.root)

    @property
    def root(self):
        return self._proj_dir

    @property
    def hash(self):
        return self.name.split('.')[-1]

    @property
    def repopath(self):
        with open(os.path.join(self.root, _FNAME.repopath), 'r', encoding='utf-8') as r:
            return r.readline().strip()

    @property
    def repo(self):
        if self._repo is None:
            from git import Repo
            self._repo = Repo(self.repopath)
        return self._repo

    @property
    def exp_dirs(self):
        fn = [os.path.join(self.root, i) for i in os.listdir(self.root)]
        fn = [f for f in fn if os.path.exists(os.path.join(f, _FNAME.repopath))]
        return fn

    @property
    def exp_names(self):
        return [os.path.basename(i) for i in self.exp_dirs]

    @property
    def exp_viewers(self):
        return [ExpViewer(i) for i in self.exp_dirs]

    def tree(self, depth=None):
        if depth == 0:
            return {
                "repopath": self.repopath,
                "name": self.name,
                "root": self.root,
                "exps": [k for k in self.exp_names]
            }
        else:
            if depth is not None:
                depth -= 1

            return {
                "repopath": self.repopath,
                "name": self.name,
                "root": self.root,
                "exps": {
                    os.path.basename(k): ExpViewer(k).tree(depth)
                    for k in self.exp_dirs
                }
            }

    def summary(self):
        print('Repo : {}'.format(self.repopath))
        print('Project Dir : {}'.format(self.root))
        print('Experiments :')
        for k in self.exp_dirs:
            print('  {}'.format(k))


class ExpViewer(_Viewer):
    """
    :param trainer_cls: cls of trainer or str
    """

    def __init__(self, exp_dir: str):
        self._exp_dir = exp_dir
        self._repo = None
        self._name = os.path.basename(self._exp_dir)

    def __iter__(self):
        for f in self.test_dirs:
            yield TestViewer(f)

    def __repr__(self):
        return "ExpViewer({} tests)".format(len(self.test_dirs))

    @property
    def name(self):
        return self._name

    @property
    def root(self):
        return self._exp_dir

    @property
    def repopath(self):
        with open(os.path.join(self.root, _FNAME.repopath), encoding='utf-8') as r:
            return r.readline().strip()

    @property
    def repo(self):
        if self._repo is None:
            from git import Repo
            self._repo = Repo(self.repopath)
        return self._repo

    @property
    def test_names(self):
        return [os.path.basename(i) for i in self.test_dirs]

    @property
    def test_dirs(self):
        if not os.path.exists(self.root):
            return []
        fs = [os.path.join(self.root, f) for f in os.listdir(self.root)]
        fs = [f for f in fs if os.path.exists(os.path.join(f, _FNAME.info))]
        fs = sorted(fs, key=lambda f: os.path.getmtime(f))
        return fs

    @property
    def test_viewers(self):
        """默认返回所有可见的 试验 的 TestViewer 对象"""
        return list(iter(self))

    @property
    def proj_name(self):
        return os.path.basename(self.proj_dir)

    @property
    def proj_dir(self):
        return os.path.dirname(self.root)

    @property
    def projviewer(self):
        return ProjViewer(self.proj_dir)

    def tree(self, depth=None):
        if depth == 0:
            return {
                "repopath": self.repopath,
                "name": self.name,
                "root": self.root,
                'tests': self.test_names
            }
        return {
            "repopath": self.repopath,
            "name": self.name,
            "root": self.root,
            'tests': {
                os.path.basename(k): TestViewer(k).tree() for k in self.test_dirs
            }
        }


class TestViewer(_Viewer):

    def __init__(self, test_dir: str):
        self._test_dir = test_dir
        self._repo = None
        self._commit = None
        self.count = int(os.path.basename(test_dir).split(".")[0])
        self._json_info = None

    def __repr__(self):
        return "Test({})".format(self.name)

    """base variables"""

    @property
    def name(self) -> str:
        return self.json_info[_INFOJ.test_name]

    @property
    def root(self):
        return self._test_dir

    @property
    def repopath(self) -> str:
        return self.json_info[_INFOJ.repo]

    """git varialbes"""

    @property
    def repo(self):
        if self._repo is None:
            from git import Repo
            self._repo = Repo(self.repopath)
        return self._repo

    @property
    def commit(self):
        if self._commit is None:
            from gitdb.util import hex_to_bin
            from git import Commit
            self._commit = Commit(self.repo, hex_to_bin(self.commit_hash))
        return self._commit

    """parent varialbes"""

    @property
    def exp_dir(self):
        return os.path.dirname(self.root)

    @property
    def expviewer(self):
        return ExpViewer(self.exp_dir)

    @property
    def test_order_count(self) -> int:
        return int(self.name.split('.')[0])

    """file infomation"""

    @property
    def fn_global_info(self) -> str:
        return os.path.join(self.root, _FNAME.info)

    """test base infomation"""

    @property
    def json_info(self) -> dict:
        if self._json_info is None:
            if not os.path.exists(self.fn_global_info):
                self._json_info = attr()
                res = self._json_info
            with open(self.fn_global_info, encoding='utf-8') as r:
                try:
                    self._json_info = attr(json.load(r))
                    res = self._json_info
                except json.JSONDecodeError as je:
                    print('Error when decoding test {}, path = {}'.format(self.name, self.root))
                    res = attr()

            res["test_dir"] = self.root
            res['visible'] = self.visible
            res['fav'] = self.isfav
            res['states'] = self.states
            return res.jsonify()
        return self._json_info.jsonify()

    @property
    def argv(self):
        return self.json_info[_INFOJ.argv]

    @property
    def commit_hash(self) -> str:
        return self.json_info[_INFOJ.commit_hash]

    @property
    def short_hash(self) -> str:
        return self.json_info[_INFOJ.short_hash]

    @property
    def success_exit(self) -> bool:
        return _INFOJ.end_code in self.json_info and self.json_info[_INFOJ.end_code] == 0

    @property
    def start_time(self):
        return date_from_str(self.json_info[_INFOJ.start_time], self.json_info[_INFOJ.time_fmt])

    @property
    def end_time(self):
        return date_from_str(self.json_info[_INFOJ.end_time], self.json_info[_INFOJ.time_fmt])

    @property
    def duration(self):
        return self.end_time - self.start_time

    """base plugins/tags operation"""

    @property
    def plugins(self) -> dict:
        dic = self.json_info[_INFOJ.plugins]
        for tag, plugins in self.json_info[_INFOJ.tags].items():
            for plugin in plugins:
                if plugin not in dic:
                    dic[plugin] = {
                        _INFOJ._tags: []
                    }
                else:
                    lis = dic[plugin].setdefault(_INFOJ._tags, [])
                    lis.append(tag)
        return dic

    @property
    def tags(self):
        dic = dict()
        for tag, plugins in self.json_info[_INFOJ.tags].items():
            if len(plugins) == 0:
                plugins = [_INFOJ._]
            for plugin in plugins:
                lis = dic.setdefault(plugin, [])
                lis.append(tag)
        return dic

    def has_tag(self, tag: str):
        return _INFOJ.tags in self.json_info and tag in self.json_info[_INFOJ.tags]

    def has_dir(self, dir: str):
        return _INFOJ.dirs in self.json_info and \
               dir in self.json_info[_INFOJ.dirs] and \
               os.path.exists(os.path.join(self.root, dir))

    def has_plugin(self, plugin: str):
        return _INFOJ.plugins in self.json_info and plugin in self.json_info[_INFOJ.plugins]

    def get_plugin(self, plugin: str):
        if self.has_plugin(plugin):
            return self.json_info[_INFOJ.plugins][plugin]
        return None

    """params CRUD"""

    @property
    def params_fn(self):
        if self.has_plugin(_BUILTIN_PLUGIN.params):
            return os.path.join(self.root, _FNAME.params)

        return None

    @property
    @lru_cache()
    def params(self):
        from ..frame.params import BaseParams
        if self.has_plugin(_BUILTIN_PLUGIN.params):
            return BaseParams().from_json(self.params_fn)
        return None

    """tensorboard CRUD"""

    @property
    def board_logdir(self):
        if not self.has_board():
            return None
        writer = self.get_plugin(_BUILTIN_PLUGIN.writer)
        return writer[_PLUGIN_KEY.WRITER.log_dir]

    @property
    @lru_cache()
    def board_reader(self) -> BoardReader:

        logdir = self.board_logdir
        if logdir is None:
            return None

        fs = os.listdir(logdir)
        fs = [i for i in fs if i.endswith('.bd') and i.startswith('events.out.tfevents')]
        if len(fs) == 0:
            return None
        if len(fs) > 1:
            import warnings
            warnings.warn('found multi tfevents file, return the first one')
        file = os.path.join(logdir, fs[0])

        return BoardReader(file)

    def tensorboard(self):
        import subprocess
        with subprocess.Popen(['tensorboard', '--logdir={}'.format(self.board_logdir)], env=dict(os.environ)) as p:
            try:
                return p.wait(timeout=None)
            except:  # Including KeyboardInterrupt, wait handled that.
                p.kill()
                raise

    def has_board(self):
        return self.has_plugin(_BUILTIN_PLUGIN.writer) and self.has_dir(_PLUGIN_DIRNAME.writer)

    """logger CRUD"""

    @property
    def log_fn(self):
        if self.has_plugin(_BUILTIN_PLUGIN.logger):
            return self.get_plugin(_BUILTIN_PLUGIN.logger)[_PLUGIN_KEY.LOGGER.fn]
        return None

    def has_log(self):
        plugin_info = self.get_plugin(_BUILTIN_PLUGIN.logger)
        if plugin_info is not None:
            return os.path.exists(plugin_info[_PLUGIN_KEY.LOGGER.fn])

    """saver CRUD"""

    @property
    def saver_dir_fn(self):
        if self.has_plugin(_BUILTIN_PLUGIN.saver):
            return self.get_plugin(_BUILTIN_PLUGIN.saver)[_PLUGIN_KEY.SAVER.ckpt_dir]
        return None

    def delete_keypoints(self):
        dir = self.saver_dir_fn
        if dir is None:
            return
        fn = [os.path.join(dir, i) for i in os.listdir(dir) if i.endswith('.ckpt') and i.startswith('keep')]
        for f in fn:
            os.remove(f)
            jfn = '{}.json'.format(f)
            if os.path.exists(jfn):
                os.remove(jfn)

    def delete_checkpoints(self):
        dir = self.saver_dir_fn
        if dir is None:
            return
        fn = [os.path.join(dir, i) for i in os.listdir(dir) if i.endswith('.ckpt') and not i.startswith('keep')]
        for f in fn:
            os.remove(f)
            jfn = '{}.json'.format(f)
            if os.path.exists(jfn):
                os.remove(jfn)

    def delete_modules(self):
        dir = self.saver_dir_fn
        if dir is None:
            return
        fn = [os.path.join(dir, i) for i in os.listdir(dir) if i.startswith('model') and i.endswith('pth')]
        for f in fn:
            os.remove(f)
            jfn = '{}.json'.format(f)
            if os.path.exists(jfn):
                os.remove(jfn)

    def delete_saver_dir(self):
        dir = self.saver_dir_fn
        if dir is None:
            return
        shutil.rmtree(dir)

    def get_tag(self, tag: str):
        if self.has_tag(tag):
            return self.json_info[_INFOJ.tags][tag]
        return None

    """states CRUD"""

    def _build_state_fn(self, state_name: str):
        return '..{}'.format(state_name)

    @property
    def isfav(self):
        return self.has_state(_TEST_BUILTIN_STATE.fav)

    @property
    def visible(self):
        return not self.has_state(_TEST_BUILTIN_STATE.hide)

    @property
    def states(self) -> Dict[str, bool]:
        fn = [i for i in os.listdir(self.root) if i.startswith('..')]
        res = {}
        for f in fn:
            with open(os.path.join(self.root, f), 'r') as r:
                res[f] = bool(int(r.readline().strip()))
        return res

    def has_state(self, state_name: str) -> bool:
        fn = os.path.join(self.root, self._build_state_fn(state_name))
        if os.path.exists(fn):
            with open(fn, 'r') as w:
                state = int(w.readline().strip())
        else:
            state = 0
        return state == 1

    def toggle_state(self, state_name: str, toggle=None) -> bool:
        fn = os.path.join(self.root, self._build_state_fn(state_name))
        if os.path.exists(fn):
            with open(fn, 'r') as r:
                state = int(r.readline().strip())
            if toggle is None:
                new_state = 1 - state
            else:
                new_state = int(toggle)
        else:
            new_state = 1

        with open(fn, 'w') as w:
            w.write('{}'.format(new_state))

        return bool(new_state)

    def mark(self, state_name: str):
        self.toggle_state(state_name, True)

    def unmark(self, state_name: str):
        self.toggle_state(state_name, False)

    def hide(self):
        self.toggle_state(_TEST_BUILTIN_STATE.hide, True)

    def show(self):
        self.toggle_state(_TEST_BUILTIN_STATE.hide, False)

    def fav(self):
        self.toggle_state(_TEST_BUILTIN_STATE.fav, True)

    def unfav(self):
        self.toggle_state(_TEST_BUILTIN_STATE.fav, False)

    """git operation"""

    def reset(self) -> Experiment:
        """
        将工作目录中的文件恢复到某个commit
        恢复快照的 git 流程:
            git branch experiment
            git add . & git commit -m ... // 保证文件最新，防止冲突报错，这一步由 Experiment() 代为完成
            git checkout <commit-id> // 恢复文件到 <commit-id>
            git checkout -b reset // 将当前状态附到新的临时分支 reset 上
            git branch experiment // 切换回 experiment 分支
            git add . & git commit -m ... // 将当前状态重新提交到最新
                // 此时experiment 中最新的commit 为恢复的<commit-id>
            git branch -D reset  // 删除临时分支
            git branch master // 最终回到原来分支，保证除文件变动外git状态完好
        Returns:
            An Experiment represents this reset operation
        """
        commit = self.commit

        old_path = os.getcwd()
        os.chdir(commit.tree.abspath)
        exp = Experiment('Reset')

        repo = self.repo
        from thexp.utils.repository import branch
        with branch(commit.repo, _GITKEY.thexp_branch) as new_branch:
            repo.git.checkout(commit.hexsha)
            repo.git.checkout('-b', 'reset')
            repo.head.reference = new_branch
            repo.git.add('.')
            ncommit = repo.index.commit("Reset from {}".format(commit.hexsha))
            repo.git.branch('-d', 'reset')
        exp.add_plugin('reset', {
            'test_name': self.name,  # 从哪个状态恢复
            'from': exp.commit.hexsha,  # reset 运行时的快照
            'where': commit.hexsha,  # 恢复到哪一次 commit，是恢复前的保存的状态
            'to': ncommit.hexsha,  # 对恢复后的状态再次进行提交，此时 from 和 to 两次提交状态应该完全相同
        })

        exp.end()
        os.chdir(old_path)
        return exp

    def archive(self) -> Experiment:
        """
        将某次 test 对应 commit 的文件打包，相关命令为
            git archive -o <filename> <commit-hash>

        Returns:
            An Experiment represents this archive operation
        """
        commit = self.commit

        old_path = os.getcwd()
        os.chdir(commit.tree.abspath)
        exp = Experiment('Archive')

        revert_path = exp.makedir('archive')
        revert_fn = os.path.join(revert_path, "code.zip")
        exp.add_plugin('archive', {'file': revert_fn,
                                   'test_name': self.name})
        with open(revert_fn, 'wb') as w:
            self.repo.archive(w, commit)

        exp.end()
        os.chdir(old_path)
        return exp

    def tree(self, _=None):
        return self.json_info

    """for representation"""

    def df_columns(self):
        return [
            "success_exit",
            "start_time",
            "end_time",
            "argv",
            "project",
            "repo",
            "exp_name",
            "test_dir",
            "plugins",
            "short_hash",
        ]

    def df_info(self):
        res = list()
        json_info = self.json_info
        res.append(self.success_exit)
        res.append(json_info.get("start_time", None))
        res.append(json_info.get("end_time", None))
        argv = list(json_info.get("argv"))
        argv[0] = os.path.relpath(argv[0], json_info.get("repo"))

        res.append(' '.join(argv))
        res.append(json_info.get("project_name", None))
        res.append(json_info.get("repo"))
        res.append(json_info.get("exp_name", None))
        res.append(json_info.get("test_dir", None))
        res.append(', '.join(list(self.plugins.keys())))

        res.append(json_info.get("short_hash", None))

        return res



from thexp import Params
from thexp.frame import BaseParams
import os
import pickle


def create_params():
    params = BaseParams()
    params.optim = params.create_optim('SGD', lr=0.1, momentum=0.9)
    params.sche = params.SCHE.Cos()
    params.schelis = params.SCHE.List(
        [params.SCHE.Cos(left=0, right=6),
         params.SCHE.Linear(left=6, right=10, start=1), ]
    )
    params.dataset = params.choice('dataset', 'cifar10', 'cifar100', 'svhn', 'stl10')
    params.n_classes = 10
    params.margin = params.arange('margin', 1.4, 0.5, 2)

    params.dataset = 'cifar100'
    params.new2 = 5
    return params


def test_baseparams():
    params = create_params()

    params.to_json('tmp')
    params2 = BaseParams()
    params2.from_json('tmp')
    os.remove('tmp')

    assert params == params2

    assert params._repeat == None
    for i, subp in enumerate(params.grid_range(5)):
        assert subp._repeat == i
    assert params._repeat == None

    iter_dataset = ['cifar10', 'cifar100']
    iter_lr = [0.3, 0.1, 0.03, 0.01]
    for i, subp in enumerate(params.grid_search('dataset', iter_dataset)):
        for j, subpp in enumerate(subp.grid_search('optim.args.lr', iter_lr)):
            assert subpp.dataset == iter_dataset[i]
            assert subpp.optim.args.lr == iter_lr[j]

    assert params == params2

    assert params.optim.args.lr == 0.1

    params.new1 = params.default(5)
    assert params.new1 == 5

    params.new2 = params.default(6)
    assert params.new2 == 5

    assert 'new3' not in params
    with open('a.pkl', 'wb') as w:
        pickle.dump(params._param_dict, w)
    with open('a.pkl', 'rb') as r:
        _param_dict = pickle.load(r)
    assert params._param_dict == _param_dict

    with open('b.pkl', 'wb') as w:
        pickle.dump(params, w)
    with open('b.pkl', 'rb') as r:
        params3 = pickle.load(r)
    assert params == params3


def test_params():
    pass



import atexit
import json
import os
import re
import sys
import warnings
from collections import namedtuple

from thexp.globals import _CONFIGL, _GITKEY, _FNAME, _OS_ENV, _PLUGIN_KEY, _BUILTIN_PLUGIN, _PLUGIN_DIRNAME
from thexp.utils.dates import curent_date

from .configs import globs

test_dir_pt = re.compile("[0-9]{4}\.[a-z0-9]{7}")
exception = namedtuple('exception_status', ['exc_type', 'exc_value', 'exc_tb'])


class Experiment:
    """
    用于目录管理和常量配置管理
    """
    user_level = _CONFIGL.running
    exp_level = _CONFIGL.globals
    repo_level = _CONFIGL.repository
    count = 0
    whole_tests = []

    def __init__(self, exp_name):
        self._start_time = None
        self._time_fmt = "%Y-%m-%d %H:%M:%S"

        self._exp_name = exp_name
        self._exp_dir = None
        self._test_dir = None
        self._project_dir = None

        self._repo = None
        self._commit = None

        self._end_state = False  # 是否试验已结束
        self._exc_dict = None  # type:exception

        self._config = globs

        self._hold_dirs = []
        self._tags = {}
        self._plugins = {}

        self._in_main = True
        self._initial()

    def __getstate__(self):
        res = {
            '_start_time': self._start_time,
            '_time_fmt': self._time_fmt,

            '_exp_name': self._exp_name,
            '_exp_dir': self.exp_dir,
            '_test_dir': self.test_dir,
            '_project_dir': self.project_dir,

            '_hold_dirs': self._hold_dirs,
            '_tags': self._tags,
            '_plugins': self._plugins,

            '_commit': self.commit.hexsha,

            '_exc_dict': self._exc_dict,
            '_end_state': self._end_state,

            '_in_main': self._in_main,
        }
        emp_ks = [k for k in res if res[k] is None]
        for k in emp_ks:
            res.pop(k)
        return res

    def __setstate__(self, state: dict):
        from gitdb.util import hex_to_bin
        from git import Commit
        self._start_time = state['_start_time']
        self._time_fmt = state['_time_fmt']

        self._exp_name = state['_exp_name']
        self._exp_dir = state['_exp_dir']
        self._test_dir = state['_test_dir']
        self._project_dir = state['_project_dir']

        self._hold_dirs = state['_hold_dirs']
        self._plugins = state['_plugins']
        self._tags = state['_tags']

        self._exc_dict = state.get('_exc_dict', None)
        self._end_state = state['_end_state']

        self._in_main = state['_in_main']

        self._repo = None
        self._commit = Commit(self.repo, hex_to_bin(state['_commit']))
        self._config = globs

    def disable_write(self):
        self._in_main = False

    def __getitem__(self, item):
        return self._config[item]

    def __setitem__(self, key, value):
        self._config[key] = value

    def _initial(self):
        """
        初始化，会在实例被创建完成后调用

        操作：
            注册 系统退出 回调
            注册 repository
            注册 实验
            提交
        """
        from thexp.utils.repository import commit as git_commit
        if self._start_time is None:
            self._start_time = curent_date(self._time_fmt)
            sys.excepthook = self.exc_end
            atexit.register(self.end)

            self._regist_repo()
            self._regist_exp()

            # 目录下所有thexp项目
            if _OS_ENV.THEXP_COMMIT_DISABLE not in os.environ:
                git_commit(self.repo, _GITKEY.commit_key, branch_name=_GITKEY.thexp_branch)

            self._write()
        else:
            warnings.warn("start the same experiment twice is not suggested.")

    def _regist_repo(self):
        """
        在相应位置注册当前实验保存的项目级别的路径（project_dir）与该repo的相应位置，便于回溯。

        共在三个位置添加相应关系（每次都会确认，未添加会添加，添加则覆盖）

        :param exp_repo:  Exprepo对象
        :return:
        """

        # 在全局添加 项目下实验存储路径与相应repo的关系，由此可以存储所有的实验存储目录
        from thexp.utils.paths import home_dir
        global_repofn = os.path.join(home_dir(), _FNAME.repo)
        if os.path.exists(global_repofn):
            with open(global_repofn, 'r', encoding='utf-8') as r:
                res = json.load(r)
        else:
            res = {}
        res[self.project_dir] = self.repo.working_dir
        with open(global_repofn, 'w', encoding='utf-8') as w:
            json.dump(res, w)

        # 在项目级别的实验存储目录下写下相应 repo ，用于方便获取
        local_repofn = os.path.join(self.project_dir, _FNAME.repopath)
        with open(local_repofn, 'w', encoding='utf-8') as w:
            w.write(self.repo.working_dir)

        # 在repo目录下记录所用过的实验存储路径到 .expsdirs，方便回溯
        repo_expdirfn = os.path.join(self.repo.working_dir, _FNAME.expsdirs)
        if os.path.exists(repo_expdirfn):
            with open(repo_expdirfn, 'r', encoding='utf-8') as r:
                _expsdirs = [i.strip() for i in r.readlines()]
        else:
            _expsdirs = []

        if self.expsdir not in _expsdirs:
            _expsdirs.append(self.expsdir)
            with open(repo_expdirfn, 'w', encoding='utf-8') as w:
                w.write("\n".join(_expsdirs))

    def _regist_exp(self):
        """
        在实验目录下存储该实验对应的repo

        Returns:
            返回是否修改，如有修改，可能需要重新提交保证最新
        """
        # 在 exp_dir 下存储对应的 repo 路径
        local_repofn = os.path.join(self.exp_dir, _FNAME.repopath)
        with open(local_repofn, 'w', encoding='utf-8') as w:
            w.write(self.repo.working_dir)

    def _write(self, **extra):
        """将试验状态写入试验目录中的 info.json """
        if self._end_state:
            return
        if not self._in_main:
            return
        res = dict(
            repo=self.repo.working_dir,
            argv=sys.argv,
            exp_name=self._exp_name,
            exp_dir=self.exp_dir,
            test_name=os.path.basename(self.test_dir),
            test_dir=self.test_dir,
            root_dir=self.expsdir,
            project_name=self.project_name,
            project_iname=os.path.basename(self.project_dir),
            project_dir=self.project_dir,
            commit_hash=self.commit_hash,
            short_hash=self.test_hash,
            dirs=self._hold_dirs,
            time_fmt=self._time_fmt,
            start_time=self._start_time,
            tags=self._tags,
            plugins=self._plugins,
            **extra,
        )
        with open(self.test_info_fn, 'w', encoding='utf-8') as w:
            json.dump(res, w, indent=2)

    @property
    def repo(self):
        from ..utils.repository import load_repo
        if self._repo is None:
            self._repo = load_repo()
        return self._repo

    @property
    def commit(self):
        from thexp.utils.repository import commit as git_commit
        if self._commit is None:
            self._commit = git_commit(self.repo, _GITKEY.commit_key, branch_name=_GITKEY.thexp_branch)
        return self._commit

    @property
    def test_hash(self) -> str:
        if self.commit is None:
            return ""

        return self.commit.hexsha[:8]

    @property
    def test_name(self) -> str:
        return os.path.basename(self.test_dir)

    @property
    def commit_hash(self) -> str:
        if self.commit is None:
            return ""
        return self.commit.hexsha

    @property
    def project_name(self) -> str:
        return self[_GITKEY.projname]

    @property
    def projkey(self) -> str:
        # return self[_CONFIGL]
        return '{}.{}'.format(self.project_name, self[_GITKEY.uuid])

    @property
    def expsdir(self) -> str:
        """
        配置中存储的，可以用于所有项目实验日志记录的根路径
        Returns:

        """
        return self[_GITKEY.expsdir].rstrip('\\/')

    @property
    def project_dir(self) -> str:
        """
        实验根目录下，项目级别的目录
        Returns:

        """
        if self._project_dir is None:
            project_dir_ = self.projkey
            self._project_dir = os.path.join(self.expsdir, project_dir_)
            os.makedirs(self._project_dir, exist_ok=True)

        return self._project_dir

    @property
    def exp_dir(self) -> str:
        """
        在每个项目下的每个实验的目录
        Returns:

        """
        if self._exp_dir is None:
            self._exp_dir = os.path.join(self.project_dir, self._exp_name)
            os.makedirs(self._exp_dir, exist_ok=True)
            with open(os.path.join(self._exp_dir, _FNAME.repopath), 'w', encoding='utf-8') as w:
                w.write(self.repo.working_dir)
        return self._exp_dir

    @property
    def test_dir(self) -> str:
        """
        获取当前 exp 目录下的 test_dir
        命名方式： 通过 序号.hash 的方式命名每一次实验
        会进行一系列判断保证在硬盘中的任何位置不会出现同名试验
        其中，hash 保证在不同时间运行的试验文件名不同，序号保证在同一进程运行的多次试验文件名不同
        {:04d}.{hash}

        Returns: 一个全局唯一的test_dir，(绝对路径)

        Notes:
            命名时还会确保存储路径下不会存在相同的序号（在生成时如果有相同序号则一直+1），方法为
            第一次获取试验目录时记录生成的序号，之后所有的序号都在此基础上生成，但如果 exp 目录变更，则
        """
        if self._test_dir is None:
            fs = os.listdir(self.exp_dir)

            if Experiment.count != 0:  # 如果
                i = self.count + 1
            else:
                atexit.register(final_report)
                i = len([f for f in fs if re.search(test_dir_pt, f) is not None]) + 1

            cf_set = {f.split('.')[0] for f in fs}  # 首位元素永远存在，不需要判断其他文件
            while "{:04d}".format(i) in cf_set:
                i += 1

            self._test_dir = os.path.join(self.exp_dir, "{:04d}.{}".format(i, self.test_hash))
            os.makedirs(self._test_dir, exist_ok=True)
            Experiment.count = i
            Experiment.whole_tests.append(os.path.basename(self._test_dir))

        return self._test_dir

    @property
    def test_info_fn(self) -> str:
        return os.path.join(self.test_dir, _FNAME.info)

    @property
    def plugins(self) -> dict:
        return self._plugins

    def makedir(self, name):
        """
        创建 test 级别的目录
        :param name: 目录名
        :return:  返回创建目录的绝对路径
        """
        d = os.path.join(self.test_dir, name)
        os.makedirs(d, exist_ok=True)
        self._hold_dirs.append(name)
        return d

    def make_exp_dir(self, name) -> str:
        """创建 exp 级别的目录"""
        d = os.path.join(self.exp_dir, name)
        os.makedirs(d, exist_ok=True)
        return d

    def end(self, end_code=0, **extra):
        """
        手动结束试验时调用该方法可以传入其他信息写入文件，
        该方法同时还通过 atexit 注册了退出钩子，如果不调用该方法，则在程序正常退出前会自动调用该方法写入结束文件。
        通过pycharm等IDE结束时，由于其结束机制不是通过抛出 KeyboardInterrupt 异常停止的，所以该方法存在没有正常调
        用的情况，此时文件中不会记录结束用时和退出状态码，也可以做为实验失败的判断条件。

        注意，结束状态仅会调用一次，无论是正常结束（对应方法end()）还是异常结束（对应方法exc_end()），
        在调用后的任何对实验的操作均不会写入到文件中
        :param end_code: 退出状态码，表明程序是否正常结束，以免引起数据混乱
        :param extra:
        :return:
        """
        self._write(
            end_time=curent_date(self._time_fmt),
            end_code=end_code,
            **extra
        )
        self._end_state = True

    def exc_end(self, exc_type, exc_val, exc_tb):
        """
        该方法注册了异常钩子，无需手动调用，在出现异常且没有处理导致程序退出的情况时，会通过该方法记录程序失败原因，
        包括异常类型和提示，同时会将异常栈内容输出到试验目录下的 Exception 文件中
        :param exc_type:  异常传入类型
        :param exc_val:  异常传入示例
        :param exc_tb:  traceback
        :return:
        """
        import traceback
        with open(os.path.join(self.test_dir, _FNAME.Exception), 'w', encoding='utf-8') as w:
            w.write("".join(traceback.format_exception(exc_type, exc_val, exc_tb)))
        self.end(
            end_code=1,
            exc_type=traceback.format_exception_only(exc_type, exc_val)[-1].strip()
        )
        self._exc_dict = exception(exc_tb, exc_val, exc_tb)
        self._end_state = True
        traceback.print_exception(exc_type, exc_val, exc_tb)

    def add_tag(self, name, *plugins):
        """
        插件的插件，用来表示tag对应的某个插件利用另一个插件做了某些什么事情
        :param name:
        :param plugins:
        :return:
        """
        self._tags[name] = plugins
        self._write()

    def add_plugin(self, key: str, value: dict = None):
        """
        注册试验中使用的插件。


        Args:
            key: 插件名
            value: 若为空，则为空字典

        Returns:

        """
        if value is None:
            value = {}
        self._plugins[key] = value
        self._write()

    def add_params(self, params):
        """
        ```python
        from thexp import Params
        p = Params()
        exp.add_params(p)
        ```
        """
        kwargs = {
            _PLUGIN_KEY.PARAMS.param_hash: params.hash(),
        }
        self.add_plugin(_BUILTIN_PLUGIN.params, kwargs)
        params.to_json(os.path.join(self.test_dir, _FNAME.params))

    def add_logger(self, fn: str):
        """
        Examples:

        ```python
        from thexp import Logger
        logger = Logger()
        fn = logger.add_log_dir(self.experiment.test_dir)
        exp.add_logger(fn)
        ```
        """
        kwargs = {
            _PLUGIN_KEY.LOGGER.log_dir: self.test_dir,
            _PLUGIN_KEY.LOGGER.fn: fn,
        }
        self.add_plugin(_BUILTIN_PLUGIN.logger, kwargs)

    def add_board(self):
        """
        Examples:
        ```python
        from torch.utils.tensorboard import SummaryWriter

        kwargs = exp.add_board()
        res = SummaryWriter(**kwargs)
        ```
        """
        d = self.makedir(_PLUGIN_DIRNAME.writer)
        kwargs = {
            _PLUGIN_KEY.WRITER.log_dir: d,
            _PLUGIN_KEY.WRITER.filename_suffix: '.bd'
        }
        self.add_plugin(_BUILTIN_PLUGIN.writer, kwargs)
        return kwargs

    def add_saver(self):
        """
        ```python
        from thexp import Saver
        kwargs = self.experiment.add_saver()
        Saver(**kwargs)
        ```
        """
        d = self.makedir(_PLUGIN_DIRNAME.saver)

        kwargs = {
            _PLUGIN_KEY.SAVER.max_to_keep: 3,
            _PLUGIN_KEY.SAVER.ckpt_dir: d
        }
        self.add_plugin(_BUILTIN_PLUGIN.saver, kwargs)
        return kwargs

    def add_rndmanager(self):
        """
        ```python
        from thexp import RndManager
        kwargs = exp.add_rndmanager()
        return RndManager(**kwargs)
        ```
        """
        d = self.make_exp_dir(_PLUGIN_DIRNAME.rnd)

        kwargs = {
            _PLUGIN_KEY.RND.save_dir: d
        }
        self.add_plugin(_BUILTIN_PLUGIN.rnd, kwargs)
        return kwargs

    def config_items(self):
        """
        三个等级的配置文件的内容
        :return:
        """
        return self._config.items()

    def add_config(self, key, value, level=_CONFIGL.globals):

        """
        添加配置
        :param key:
        :param value:
        :param level: 配置级别，查看 globals._CONFIGL
        :return:
        """
        self._config.add_value(key, value, level)

    def regist_exit_hook(self, func):
        """
        提供给用户的额外注册退出钩子，传入的参数有两个，
        第一个参数为该 Experiment 的实例对象
        第二个参数为一个 exception_status 实例，当程序的退出是因为某异常引起时，该参数会被传入
            exception_status 是一个 namedtuple('exception_status', ['exc_type', 'exc_value', 'exc_tb'])
            可以通过 Python 内置的 traceback 库来处理，如

            import traceback
            traceback.print_exception(exc_type, exc_val, exc_tb)
        :param func: 退出前调用的钩子，
        :return:
        """
        import atexit
        def exp_func():
            func(self, self._exc_dict)

        atexit.register(exp_func)


def final_report():
    print('Test end:')
    import pprint
    pprint.pprint(Experiment.whole_tests)



"""

"""

from thexp.frame.meter import Meter, AvgMeter
import torch

from thexp import Meter, AvgMeter


def test_meter():
    meter = Meter()
    meter.a = 1
    meter.b = 0.5
    meter.c = 2.5
    meter.percent(meter.b_)
    meter.int(meter.c_)
    meter.float(meter.a_)


def test_avgmeter():
    avg = AvgMeter()

    avg.average(avg.a_)
    avg.a = 1
    avg.a = 2
    avg.a = 3
    assert avg.a.avg == 2  # AvgItem object
    assert avg.a.item == 3

    avg.b = 1
    avg.b = 2
    assert avg.b == 2  # int object

    avg.average(avg.c_)
    x, y = torch.tensor([1, 2, 3]), torch.tensor([4, 5, 6])
    avg.c = x
    avg.c = y
    assert (avg.c.avg == (x + y) / 2).all()
    assert (avg.c.item == y).all()

    assert avg.meter.a == 3
    assert avg.avg.a == 2
    assert avg.meter.b == 2
    assert avg.avg.b == 2



"""

"""
import operator as op
import sys
from functools import reduce
from typing import Union

import numpy as np


def pin(*sizes, dtype: str = "float32", format: bool = False) -> Union[int, str]:
    """
    计算相应尺寸和类型“至少”占用的内存大小
    由于除了存放相应 size 个数的数据外，还有用于辅助的各类变量，因此说是至少

    Args:
        *sizes:
        dtype:
        format:

    Returns:
        if format:
            return 格式化后的大小
        else:
            return 整数，表示 byte 数
    Examples:
    >>> print(pin(10,50000,format=True))
    """
    # len(sizes) * 16 + 80
    one = np.array(1, dtype=dtype)

    size = sys.getsizeof(one) - 80
    mem = reduce(op.mul, sizes) * size

    if format:
        unit = "B"
        if mem > 1024:
            mem /= 1024
            unit = "KB"
            if mem > 1024:
                mem /= 1024
                unit = "MB"
                if mem > 1024:
                    mem /= 1024
                    unit = "GB"
        return "{:.4f} {}".format(mem, unit)

    return mem



"""
Methods about iterator
"""
from collections.abc import Iterator


class safe_cycle():
    class _cycle_none():
        pass

    def __init__(self, iter_item):
        self.item = iter_item
        self.iter = None

    def __next__(self):
        if self.iter is None:
            self.iter = iter(self.item)
            cur = next(self.iter, safe_cycle._cycle_none())
            if isinstance(cur, safe_cycle._cycle_none):
                raise StopIteration('item has no element.')
            return cur
        else:
            cur = next(self.iter, safe_cycle._cycle_none())
            if not isinstance(cur, safe_cycle._cycle_none):
                return cur
            else:
                self.iter = iter(self.item)
                return next(self)

    def __iter__(self):
        while True:
            yield next(self)


def deep_chain(item):
    """flatten iterator"""
    if isinstance(item, Iterator):
        for i in item:
            if isinstance(i, Iterator):
                for ii in deep_chain(i):
                    yield ii
            else:
                yield i
    else:
        yield item


def is_same_type(items, ty=None):
    for item in items:
        if ty is None:
            ty = type(item)
        else:
            if type(item) != ty:
                return False
    return True



"""

"""

from .charts import Curve, Parallel
from .querys import Query, BoardQuery, TestsQuery, ExpsQuery, ReposQuery, Q
from .viewer import TestViewer, ProjViewer, ExpViewer

from .constrain import C



"""
define some global string variables in case name mistakes
"""
from thexp import __VERSION__


class _REPOJ:
    repopath = 'repopath'
    exps = 'exps'
    exp_root = 'exp_root'


class _INFOJ:
    repo = 'repo'
    argv = 'argv'
    test_name = 'test_name'
    commit_hash = 'commit_hash'
    short_hash = 'short_hash'
    dirs = 'dirs'
    time_fmt = 'time_fmt'
    start_time = 'start_time'
    tags = 'tags'
    _tags = '_tags'
    _ = '_'
    plugins = 'plugins'
    end_time = 'end_time'
    end_code = 'end_code'


class _CONFIGL:
    running = 'running'
    globals = 'globals'
    repository = 'repository'


class _GITKEY:
    thexp = 'thexp'
    projname = 'projname'
    expsdir = 'expsdir'
    uuid = 'uuid'
    section_name = 'thexp'
    thexp_branch = 'experiment'
    commit_key = 'thexp-commit'


class _FNAME:
    Exception = 'Exception'
    info = 'info.v1.json'
    repo = 'repo.v1.json'
    params = 'params.v1.json'
    repopath = '.repopath'
    expsdirs = '.expsdirs'
    gitignore = ".gitignore"
    gitignore_version = '.thexp.{}'.format(__VERSION__)


class _TEST_BUILTIN_STATE:
    hide = 'hide'
    fav = 'fav'


class _ML:
    train = 'train'
    test = 'test'
    eval = 'eval'
    cuda = 'cuda'


class _BUILTIN_PLUGIN:
    trainer = 'trainer'
    params = 'params'
    writer = 'writer'
    logger = 'logger'
    saver = 'saver'
    rnd = 'rnd'


class _PLUGIN_DIRNAME:
    writer = 'board'
    writer_tmp = 'board_tmp'
    saver = 'modules'
    rnd = 'rnd'


class _PLUGIN_KEY:
    class WRITER:
        log_dir = 'log_dir'
        filename_suffix = 'filename_suffix'
        dir_name = 'board'

    class LOGGER:
        log_dir = 'log_dir'
        fn = 'fn'

    class RND:
        save_dir = 'save_dir'

    class PARAMS:
        param_hash = 'param_hash'

    class TRAINER:
        path = 'path'
        doc = 'doc'
        fn = 'module'
        class_name = 'class_name'

    class SAVER:
        max_to_keep = 'max_to_keep'
        ckpt_dir = 'ckpt_dir'


class _INDENT:
    tab = '  '
    ttab = '    '
    tttab = '      '


class _DLEVEL:
    proj = 'proj'
    exp = 'exp'
    test = 'test'


class _OS_ENV:
    CUDA_VISIBLE_DEVICES = 'CUDA_VISIBLE_DEVICES'
    THEXP_COMMIT_DISABLE = 'THEXP_COMMIT_DISABLE'
    IGNORE_REPO = 'THEXP_IGNORE_REPO'



"""

"""
from contextlib import contextmanager


class Code():
    def __init__(self):
        self.lines = []

    def add_line(self, code):
        if isinstance(code, str):
            self.lines.append(code)
        elif isinstance(code, (list, tuple)):
            self.lines.extend(code)

    def new_line(self):
        self.lines.append("")

    def wrap(self):
        return "\n".join(self.lines)

    def __repr__(self):
        return self.wrap()


class Table():
    def __init__(self):
        self.lines = []
        self.heads = []
        self.mid = []
        self.col_len = None

    def update_col_len(self, line):
        if self.col_len is None:
            self.col_len = len(line)
        else:
            self.col_len = max(self.col_len, len(line))
            self.refresh_cols()

    def refresh_cols(self):
        if len(self.heads) != self.col_len:
            self.heads.extend(["-"] * (self.col_len - len(self.heads)))
        self.mid = [":---:"] * self.col_len
        for line in self.lines:
            if len(line) != self.col_len:
                line.extend(["-"] * (self.col_len - len(line)))

    def head(self, head: [list, tuple]):
        self.heads = head
        self.update_col_len(head)

    def append(self, line: [list, tuple]):
        self.lines.append(line)
        self.update_col_len(line)

    def mat(self):
        return [self.heads, *self.lines]

    def __repr__(self):
        res = []

        def add_line(line: list):
            res.append("|{}|\n".format("|".join([str(i) for i in line])))

        add_line(self.heads)
        add_line(self.mid)
        for line in self.lines:
            add_line(line)

        return "".join(res)


class Markdown():
    def __init__(self):
        self.lines = []

    def add_title(self, title, level=1):
        self.lines.append("\n{} {}\n".format("#" * level, title))
        return self

    def add_text(self, text):
        self.lines.append(text)
        return self

    def add_bold(self, text):
        self.lines.append(" **{}** ".format(text))
        return self

    def add_italic(self, text):
        self.lines.append(" *{}* ".format(text))
        return self

    def newline(self):
        # if len(self.lines) > 0 and len(self.lines[-1].strip()) != 0 and not self.lines[-1].endswith("\n"):
        self.lines.append("\n")
        return self

    def add_url(self, url, text):
        self.lines.append(" [{}]({}) ".format(text, url))
        return self

    def add_table(self, mat: list, head: list = None):
        with self.table() as tb:  # type:Table
            if head is None:
                tb.head(mat[0])
                mat = mat[1:]
            else:
                tb.head(head)

            for line in mat:
                tb.append(line)

    def __str__(self) -> str:
        return "".join(self.lines)

    def add_picture(self, url, text=None, inline=True):
        if inline:
            self.lines.append(" ![{}]({}) ".format(text, url))
        else:
            self.lines.append("\n\n![{}]({})\n\n".format(text, url))
        return self

    @contextmanager
    def table(self) -> Table:
        self.lines.append("\n\n")
        tb = Table()
        yield tb  # type:Table
        self.lines.append(str(tb))
        self.lines.append("\n\n")

    @contextmanager
    def parameter(self):
        self.lines.append("\n\n")
        yield self  # type:Markdown
        self.lines.append("\n\n")

    @contextmanager
    def quote(self):
        self.lines.append("\n\n")
        md = Markdown()
        yield md  # type:Markdown
        self.extends(md.wrap_quote())
        self.lines.append("\n\n")

    @contextmanager
    def code(self, lang=None) -> Code:
        """
        with md.code() as code:
            code.
        """
        self.lines.append("\n\n```{}\n".format(lang))
        code = Code()
        yield code
        self.lines.append(code.wrap())
        self.lines.append("\n```\n\n")

    def add_code(self, code, lang="", with_wrap=True):
        import textwrap
        if with_wrap:
            with self.code(lang) as cd:
                cd.add_line(textwrap.dedent(code).strip("\n"))
        else:
            self.lines.append("\n\n")
            self.lines.append(code)
            self.lines.append("\n\n")

        return self

    def extends(self, obj):
        if isinstance(obj, Markdown):
            self.lines.extend(obj.lines)
        elif isinstance(obj, (list, tuple)):
            self.lines.extend(obj)
        elif isinstance(obj, str):
            self.lines.append(obj)

        return self

    def wrap_quote(self):
        return ["> {}\n".format(i) for i in "".join(self.lines).split("\n")]

    def to_str(self):
        return str(self)

    def to_file(self, fn, mode="w"):
        with open(fn, mode, encoding="utf-8") as w:
            w.write(self.to_str())



"""

"""


def draw_dict():
    res = dict()

    res['x'] = []
    res['y'] = []
    return res


class default():
    def __init__(self, default, warn=False):
        self.default = default
        self.warn = warn



"""

Q = Query()

Q.repos()['SemiEnhance'].exps()['some_exp'].tests()[0,1,2]...
Q.repos()['SemiEnhance'].exps()['some_exp'].tests('0001','0002')...
Q.exps()[:2].tests()
Q.tests().first() -> TestViewer
Q.tests().last() -> TestViewer
Q.tests().last() -> TestViewer

    ... list all repos

blur search:
Q.rfind/efind/tfind/
Q.rfind('th')
Q.tfind('')

compare params:
Q.tests('xxx').compare_params()


compare files:
Q.tests('xxx').compare_files(include_current=True)

    output a table

Q.tests('xxx').compare().files()
Q.tests('xxx').compare().file(a='...',b=None,skip=0) #


"""
import json
import numbers

import warnings
import numpy as np
import os
from itertools import chain
from collections import OrderedDict
from ..base_classes.trickitems import NoneItem
from pprint import pformat
from typing import List, Iterator, Set
from datetime import datetime, timedelta
import pandas as pd

from thexp.base_classes.list import llist
from thexp.globals import _FNAME
from thexp.utils.paths import home_dir
from .reader import BoardReader
from .viewer import TestViewer
from ..utils.iters import is_same_type
from ..globals import _BUILTIN_PLUGIN
from .constrain import Constrain, ParamConstrain, MeterConstrain

# 虽然好像没什么用但还是设置上了
pd.set_option('display.max_colwidth', 160)
pd.set_option('colheader_justify', 'center')


class Query:

    def tests(self, *items):
        return self.repos().exps().tests(*items)

    def exps(self, *items):
        return self.repos().exps(*items)

    def repos(self, *items):
        if len(items) == 0:
            return ReposQuery(*self.projs_list)
        else:
            return ReposQuery(*self.projs_list)[items]

    @property
    def projs_list(self):
        global_repofn = os.path.join(home_dir(), _FNAME.repo)
        if not os.path.exists(global_repofn):
            res = {}
        else:
            with open(global_repofn, 'r', encoding='utf-8') as r:
                res = json.load(r)

        re_write = False

        repos = []
        projs = []
        for k, v in res.items():  # log_dir(proj level), repopath
            if not os.path.exists(k):  # 不仅需要日志记录，还需要该试验目录确实存在
                re_write = True
                continue

            repos.append(v)
            projs.append(k)

        if re_write:
            with open(global_repofn, 'w', encoding='utf-8') as w:
                json.dump({k: v for k, v in zip(projs, repos)}, w)

        return projs, repos


class ReposQuery:
    """
    以项目为单位的Query，每个项目可能对应多个实验存储路径（一般不主动更改的情况下只有一个）
    """

    def __init__(self, projs: List[str], repos: List[str]):
        self.projs = llist(projs)
        self.repos = llist(repos)

        self.proj_names = [os.path.basename(i) for i in self.projs]

    def __and__(self, other):
        if isinstance(other, ReposQuery):
            combine = set(self.projs) & set(other.projs)
            projs = []
            repos = []
            for proj, repo in chain(zip(self.projs, self.repos), zip(other.projs, other.repos)):
                if proj in combine:
                    combine.remove(proj)
                    projs.append(proj)
                    repos.append(repo)
            return ReposQuery(projs, repos)
        raise TypeError(self, other)

    def __or__(self, other):
        if isinstance(other, ReposQuery):
            combine = set(self.projs) | set(other.projs)
            projs = []
            repos = []
            for proj, repo in chain(zip(self.projs, self.repos), zip(other.projs, other.repos)):
                if proj in combine:
                    combine.remove(proj)
                    projs.append(proj)
                    repos.append(repo)
            return ReposQuery(projs, repos)
        raise TypeError(self, other)

    def __getitem__(self, items):
        res = []
        if isinstance(items, (Iterator, list, tuple)):
            assert is_same_type(items)
            for item in items:
                if isinstance(item, str):
                    try:
                        idx = self.proj_names.index(item)
                        res.append(idx)
                    except:
                        raise IndexError(item)
                elif isinstance(item, int):
                    res.append(item)
            return ReposQuery(self.projs[res], self.repos[res])  # do not have type error
        elif isinstance(items, int):
            res.append(items)
            return ReposQuery(self.projs[res], self.repos[res])  # do not have type error
        elif isinstance(items, str):
            return self.__getitem__([items])
        elif isinstance(items, slice):
            return ReposQuery(self.projs[items], self.repos[items])  # do not have type error

    def __repr__(self):
        return self.df().__repr__()

    __str__ = __repr__

    def __len__(self):
        return len(self.repos)

    def _repr_html_(self):
        with pd.option_context('display.max_colwidth', 0):
            with pd.option_context('colheader_justify', 'center'):
                return self.df()._repr_html_()

    def df(self):
        res = []
        for repo, proj in zip(self.repos, self.projs):
            res.append((os.path.basename(proj), repo))
        return pd.DataFrame(res, columns=['name', 'Repo path'])

    def tests(self, *items):
        return self.exps().tests(*items)

    def exps(self, *items):
        if len(items) == 0:
            viewers = self.to_viewers()
            exp_dirs = []
            for viewer in viewers:
                exp_dirs.extend(viewer.exp_dirs)
            return ExpsQuery(exp_dirs)
        else:
            return self.exps()[items]

    @property
    def empty(self):
        return len(self.repos) == 0

    @property
    def isitem(self):
        return len(self.repos) == 1

    def to_viewer(self):
        if self.isitem:
            from .viewer import ProjViewer
            return ProjViewer(self.projs[0])
        raise ValueError('only one projs contains can be converted to ProjViewer')

    def to_viewers(self):
        from .viewer import ProjViewer
        return [ProjViewer(i) for i in self.projs]

    def delete(self):
        for viewer in self.to_viewers():
            viewer.delete()


class ExpsQuery:
    def __init__(self, exp_dirs):
        self.exp_dirs = llist(exp_dirs)
        self.exp_names = [os.path.basename(i) for i in self.exp_dirs]

    def __and__(self, other):
        if isinstance(other, ExpsQuery):
            combine = set(self.exp_dirs) & set(other.exp_dirs)
            exp_dirs = []
            for exp_dir in chain(self.exp_dirs, other.exp_dirs):
                if exp_dir in combine:
                    combine.remove(exp_dir)
                    exp_dirs.append(exp_dir)
            return ExpsQuery(exp_dirs)
        raise TypeError(self, other)

    def __or__(self, other):
        if isinstance(other, ExpsQuery):
            combine = set(self.exp_dirs) | set(other.exp_dirs)
            exp_dirs = []
            for exp_dir in chain(self.exp_dirs, other.exp_dirs):
                if exp_dir in combine:
                    combine.remove(exp_dir)
                    exp_dirs.append(exp_dir)
            return ExpsQuery(exp_dirs)
        raise TypeError(self, other)

    def __getitem__(self, items):
        res = []
        if isinstance(items, (Iterator, list, tuple)):
            if len(items) == 0:
                return self
            assert is_same_type(items)
            for i, exp in enumerate(self.exp_names):
                if isinstance(items[0], str) and exp in items:
                    res.append(i)
                elif isinstance(items[0], int) and i in items:
                    res.append(i)
            return ExpsQuery(self.exp_dirs[res])  # do not have type error
        elif isinstance(items, int):
            res.append(items)
            return ExpsQuery(self.exp_dirs[res])  # do not have type error
        elif isinstance(items, str):
            return self.__getitem__([items])
        elif isinstance(items, slice):
            return ExpsQuery(self.exp_dirs[items])  # do not have type error
        else:
            raise TypeError(items)

    def __repr__(self):
        if self.empty:
            return '[Empty]'
        return self.df().__repr__()

    __str__ = __repr__

    def __len__(self):
        return len(self.exp_names)

    def _repr_html_(self):
        if self.empty:
            return '<pre>[Empty]</pre>'
        return self.df()._repr_html_()

    def df(self):
        res = []
        names = []
        for viewer in self.to_viewers():
            res.append(viewer.test_names)
            names.append(viewer.name)
        df = pd.DataFrame(res, index=names).T
        return df

    def tests(self, *items):
        if len(items) == 0:
            viewers = self.to_viewers()
            exp_dirs = []
            for viewer in viewers:
                exp_dirs.extend(viewer.test_dirs)
            return TestsQuery(exp_dirs)
        else:
            return self.tests()[items]

    @property
    def empty(self):
        return len(self.exp_dirs) == 0

    @property
    def isitem(self):
        return len(self.exp_dirs) == 1

    def to_viewer(self):
        if self.isitem:
            from .viewer import ExpViewer
            return ExpViewer(self.exp_dirs[0])
        raise ValueError('only one projs contains can be converted to ProjViewer')

    def to_viewers(self):
        from .viewer import ExpViewer
        return [ExpViewer(i) for i in self.exp_dirs]


class BoardQuery():
    def __init__(self, board_readers: List[BoardReader], test_viewers: List[TestViewer]):
        self.board_readers = llist(board_readers)  # type:llist[BoardReader]
        self.test_viewers = llist(test_viewers)  # type:llist[TestViewer]

    def __and__(self, other):
        if isinstance(other, BoardQuery):
            combine = set(self.test_viewers) & set(other.test_viewers)
            readers = []
            test_viewers = []
            for reader, test_viewer in chain(zip(self.board_readers, self.test_viewers),
                                             zip(other.board_readers, other.test_viewers)):
                if test_viewer in combine:
                    combine.remove(test_viewer)
                    readers.append(reader)
                    test_viewers.append(test_viewer)
            return BoardQuery(readers, test_viewers)
        raise TypeError(self, other)

    def __or__(self, other):
        if isinstance(other, BoardQuery):
            combine = set(self.test_viewers) | set(other.test_viewers)
            readers = []
            test_viewers = []
            for reader, test_viewer in chain(zip(self.board_readers, self.test_viewers),
                                             zip(other.board_readers, other.test_viewers)):
                if test_viewer in combine:
                    combine.remove(test_viewer)
                    readers.append(reader)
                    test_viewers.append(test_viewer)
            return BoardQuery(readers, test_viewers)
        raise TypeError(self, other)

    def __getitem__(self, items):
        res = []
        if isinstance(items, (Iterator, list, tuple)):
            assert is_same_type(items)
            names = [i.name for i in self.test_viewers]
            for item in items:
                if isinstance(item, int):
                    res.append(item)
                elif isinstance(item, str):
                    res.append(names.index(item))
            return BoardQuery(self.board_readers[res], self.test_viewers[res])  # do not have type error
        elif isinstance(items, int):
            res.append(items)
            return BoardQuery(self.board_readers[res], self.test_viewers[res])  # do not have type error
        elif isinstance(items, str):
            return self.__getitem__([items])
        elif isinstance(items, slice):
            return BoardQuery(self.board_readers[items], self.test_viewers[items])  # do not have type error

    def __repr__(self):
        return 'BoardQuery({})'.format(pformat(self.test_viewers))

    __str__ = __repr__

    def __len__(self):
        return len(self.board_readers)

    @property
    def param_keys(self):
        res = set()
        for tv in self.test_viewers:  # type:TestViewer
            for k, v in tv.params.inner_dict.walk():
                res.add(k)
        return res

    @property
    def scalar_tags(self) -> Set[str]:
        """返回所有board reader 重合的tags"""
        from functools import reduce
        from operator import and_

        scalars_tags = [set(i.scalars_tags) for i in self.board_readers]
        return reduce(and_, scalars_tags)

    def has_scalar_tags(self, tag):
        res = []
        for i, reader in enumerate(self.board_readers):
            if tag in reader.scalars_tags:
                res.append(i)
        return self[res]

    def values(self, tag, with_step=False):
        res = []
        for reader in self.board_readers:
            try:
                val = reader.get_scalars(tag)
                if not with_step:
                    res.append(val.values)
                else:
                    res.append(val)
            except:
                res.append(None)
        return res

    def parallel_dicts(self, *constrains: Constrain):
        """
        构建 平行图的字典，返回一个字典，包含了每个 tag 对应的所有试验的N个记录

        Args:
            *constrains:
            backend:

        Returns:

        """
        tag_dict = {}

        for constrain in constrains:
            tag = constrain._name
            if isinstance(constrain, ParamConstrain):
                tag_values = list()
                for tv in self.test_viewers:  # type:TestViewer
                    params = tv.params

                    for k, v in params.inner_dict.walk():
                        # check name
                        if k != tag:
                            continue

                        # check constrain conditoin
                        if constrain._constrain != None and not constrain._constrain(v, constrain._value):
                            continue

                        if isinstance(v, (numbers.Number, str)):  # include int, float, bool
                            tag_values.append(v)

                tag_dict[tag] = tag_values

            elif isinstance(constrain, MeterConstrain):
                if constrain._name not in self.scalar_tags:
                    warnings.warn('{} can not found in this query, and will be ignored.'.format(constrain))
                    continue

                constrain_func = constrain._constrain
                values_lis = self.values(tag)
                assert all([len(values) > 0 for values in values_lis])

                if constrain_func is None:  # 自动判断约束类型
                    # 判断方式：tag中 loss 那么取最小值，有 acc 那么取最大值，否则根据初始值和末尾值的大小关系进行判断
                    if 'loss' in tag:
                        constrain_func = np.min
                    elif 'acc' in tag:
                        constrain_func = np.max
                    else:
                        for values in values_lis:
                            if values[0] > values[-1]:
                                constrain_func = np.min
                            else:
                                constrain_func = np.max
                            break

                evalue = [constrain_func(values) for values in values_lis]

                tag_dict[tag] = evalue

        return tag_dict

    def line(self, tag, backend='matplotlib'):
        """plot values of the tag of the tests selected"""
        from .charts import Curve
        figure = {}
        for bd, test_viewer in zip(self.board_readers, self.test_viewers):
            test_name = test_viewer.name

            scalars = bd.get_scalars(tag)
            figure[test_name] = {
                'name': test_name,
                'x': scalars.steps,
                'y': scalars.values,
            }
        curve = Curve(figure, title=tag)
        plot_func = getattr(curve, backend, None)
        if plot_func is None:
            raise NotImplementedError(backend)
        else:
            return plot_func()

    def tvalues(self, tags: List[str], with_step=False):
        """get tags of one test"""
        assert len(self.board_readers) == 1
        res = []
        for tag in tags:
            try:
                val = self.board_readers[0].get_scalars(tag)
                if not with_step:
                    res.append(val.values)
                else:
                    res.append(val)
            except:
                res.append(None)
        return res

    def tline(self, tags: List[str], backend='matplotlib'):
        """line of multi-tags of one test"""
        assert len(self.board_readers) == 1
        from .charts import Curve
        figure = {}
        for tag in tags:
            scalars = self.board_readers[0].get_scalars(tag)
            figure[tag] = {
                'name': tag,
                'x': scalars.steps,
                'y': scalars.values,
            }
        curve = Curve(figure, title=self.test_viewers[0].name)
        plot_func = getattr(curve, backend, None)
        if plot_func is None:
            raise NotImplementedError(backend)
        else:
            return plot_func()

    def parallel(self, *constrains: Constrain, backend='matplotlib'):
        """draw parallel with the constrains to compare tests"""
        tag_dict = self.parallel_dicts(*constrains)
        from .charts import Parallel
        parallel = Parallel([i.name for i in self.test_viewers], tag_dict)

        plot_func = getattr(parallel, backend, None)
        if plot_func is None:
            raise NotImplementedError(backend)
        else:
            return plot_func()

    def summary(self):
        from collections import defaultdict
        res = defaultdict(list)
        for br, tv in zip(self.board_readers, self.test_viewers):  # type:BoardReader,TestViewer
            for tag in br.scalars_tags:
                res[tag].append(tv.name)

        print('Scalar tags:')
        for k, v in res.items():
            if len(v) == len(self.board_readers):
                print('{} (all)'.format(k))
            else:
                print('{} ({})'.format(k, v))


class TestsQuery:

    def __init__(self, test_dirs: List[str]):
        self.test_dirs = llist(test_dirs)
        self.test_names = [os.path.basename(i) for i in self.test_dirs]

    def __and__(self, other):
        if isinstance(other, TestsQuery):
            combine = set(self.test_names) & set(other.test_names)
            test_dirs = []
            for test_dir, test_name in chain(zip(self.test_dirs, self.test_names),
                                             zip(other.test_dirs, other.test_names)):
                if test_name in combine:
                    combine.remove(test_name)
                    test_dirs.append(test_dir)
            return TestsQuery(test_dirs)
        raise TypeError(self, other)

    def __or__(self, other):
        if isinstance(other, TestsQuery):
            combine = set(self.test_names) | set(other.test_names)
            test_dirs = []
            for test_dir, test_name in chain(zip(self.test_dirs, self.test_names),
                                             zip(other.test_dirs, other.test_names)):
                if test_name in combine:
                    combine.remove(test_name)
                    test_dirs.append(test_dir)
            return TestsQuery(test_dirs)
        raise TypeError(self, other)

    def __getitem__(self, items):
        res = []
        if isinstance(items, (Iterator, list, tuple)):
            assert is_same_type(items)
            for item in items:
                if isinstance(item, str):
                    try:
                        idx = self.test_names.index(item)
                        res.append(idx)
                    except:
                        raise IndexError(item)
                elif isinstance(item, int):
                    res.append(item)
            return TestsQuery(self.test_dirs[res])  # do not have type error
        elif isinstance(items, int):
            res.append(items)
            return TestsQuery(self.test_dirs[res])  # do not have type error
        elif isinstance(items, str):
            return self.__getitem__([items])
        elif isinstance(items, slice):
            return TestsQuery(self.test_dirs[items])  # do not have type error

    def __repr__(self):
        if self.empty:
            return '[Empty]'
        return self.df().__repr__()

    def _repr_html_(self):
        if self.empty:
            return '<pre>[Empty]</pre>'
        return self.df()._repr_html_()

    __str__ = __repr__

    def __len__(self):
        return len(self.test_names)

    def df(self):
        df = pd.DataFrame([viewer.df_info() for viewer in self.to_viewers()],
                          columns=self[0].to_viewer().df_columns(),
                          index=self.test_names)
        return df

    def params_df(self):
        params_lis = [vw.params for vw in self.to_viewers()]
        res = []
        for params in params_lis:
            dic = OrderedDict()
            for k, v in params.inner_dict.walk():
                dic[k] = v
            res.append(dic)

        df = pd.DataFrame(res, index=[vw.name for vw in self.to_viewers()]).T
        return df

    def boards(self):
        from thexp.base_classes.errors import NoneWarning
        viewers = []
        boards = []
        for vw in self.to_viewers():
            if vw.board_reader is None:
                warnings.warn('{} have no board, they will be removed.'.format(vw.name), NoneWarning)
                continue
            viewers.append(vw)
            boards.append(vw.board_reader)
        return BoardQuery(boards, viewers)

    @property
    def empty(self):
        return len(self.test_dirs) == 0

    @property
    def isitem(self):
        return len(self.test_dirs) == 1

    def to_viewers(self):
        """convert testquery to  testviewer list"""
        from .viewer import TestViewer
        return [TestViewer(i) for i in self.test_dirs]

    def to_viewer(self):
        """convert testquery to single testviewer if there is only one test in queryset"""
        if self.isitem:
            from .viewer import TestViewer
            return TestViewer(self.test_dirs[0])
        raise ValueError('only one test_dirs contains can be converted to TestViewer')

    """filter tests"""

    def first(self):
        return self[0]

    def head(self, num=5):
        return self[:5]

    def last(self):
        return self[-1]

    def tail(self, num=5):
        return self[-5:]

    def time_range(self, left_time=None, right_time=None):
        """
        筛选 start_time 位于某时间区间内的 test
        Args:
            left_time:
            right_time:

        Returns:

        """
        if left_time is None and right_time is None:
            return self

        res = []
        for i, viewer in enumerate(self.to_viewers()):
            if left_time is None:
                if viewer.start_time < right_time:
                    res.append(i)
            elif right_time is None:
                if viewer.start_time > left_time:
                    res.append(i)
            else:
                if viewer.start_time > left_time and viewer.start_time < left_time:
                    res.append(i)

        return self[res]

    def in_time(self,
                minutes=0,
                hours=0,
                days=0,
                weeks=0,
                seconds=0):
        delta = timedelta(minutes=minutes, days=days, hours=hours, weeks=weeks, seconds=seconds)
        return self.time_range(left_time=datetime.now() - delta)

    def success(self):
        res = []
        for i, viewer in enumerate(self.to_viewers()):
            if viewer.success_exit:
                res.append(i)
        return self[res]

    def failed(self):
        res = []
        for i, viewer in enumerate(self.to_viewers()):
            if not viewer.success_exit:
                res.append(i)
        return self[res]

    def has_tag(self, tag: str, toggle=True):
        res = []
        for i, viewer in enumerate(self.to_viewers()):
            if viewer.has_tag(tag):
                if toggle:
                    res.append(i)
            else:
                if not toggle:
                    res.append(i)
        return self[res]

    def has_plugin(self, plugin, toggle=True):
        res = []
        for i, viewer in enumerate(self.to_viewers()):
            if viewer.has_plugin(plugin):
                if toggle:
                    res.append(i)
            else:
                if not toggle:
                    res.append(i)

        return self[res]

    def has_board(self, toggle=True):
        return self.has_plugin(_BUILTIN_PLUGIN.writer, toggle)

    def has_log(self, toggle=True):
        return self.has_plugin(_BUILTIN_PLUGIN.logger, toggle)

    def has_saver(self, toggle=True):
        return self.has_plugin(_BUILTIN_PLUGIN.saver, toggle)

    def has_params(self, toggle=True):
        return self.has_plugin(_BUILTIN_PLUGIN.params, toggle)

    def has_trainer(self, toggle=True):
        return self.has_plugin(_BUILTIN_PLUGIN.trainer, toggle)

    def has_dir(self, dir: str, toggle=True):
        res = []
        for i, viewer in enumerate(self.to_viewers()):
            if viewer.has_dir(dir):
                if toggle: res.append(i)
            else:
                if not toggle: res.append(i)
        return self[res]

    def has_state(self, state_name: str, toggle=True):
        res = []
        for i, viewer in enumerate(self.to_viewers()):
            if viewer.has_state(state_name):
                if toggle: res.append(i)
            else:
                if not toggle: res.append(i)
        return self[res]

    def filter_params(self, *constrains: ParamConstrain):
        """
        Query test with param constrains.

        Args:
            *constrains: a ParamConstrain instance.

        Returns:

        """
        if len(constrains) == 0:
            return self

        res = []
        params_lis = [vw.params for vw in self.to_viewers()]
        for i, params in enumerate(params_lis):
            feasible = True
            for constrain in constrains:
                if params is None:
                    feasible = False
                    break

                left = params.get(constrain._name, NoneItem())
                if isinstance(left, NoneItem) and not constrain._allow_none:
                    feasible = False
                    break

                right = constrain._value
                if not constrain._constrain(left, right):
                    feasible = False
                    break
            if feasible:
                res.append(i)

        return self[res]

    """update test state"""

    def delete(self):
        for viewer in self.to_viewers():
            viewer.delete()

    def delete_modules(self):
        for viewer in self.to_viewers():
            viewer.delete_modules()

    def delete_keypoints(self):
        for viewer in self.to_viewers():
            viewer.delete_keypoints()

    def delete_checkpoints(self):
        for viewer in self.to_viewers():
            viewer.delete_checkpoints()

    def toggle_state(self, state_name: str, toggle=None):
        for viewer in self.to_viewers():
            viewer.toggle_state(state_name, toggle)

    def mark(self, state_name: str):
        self.toggle_state(state_name, True)

    def unmark(self, state_name: str):
        self.toggle_state(state_name, False)

    def hide(self):
        for viewer in self.to_viewers():
            viewer.hide()

    def show(self):
        for viewer in self.to_viewers():
            viewer.show()

    def fav(self):
        for viewer in self.to_viewers():
            viewer.fav()

    def unfav(self):
        for viewer in self.to_viewers():
            viewer.unfav()


Q = Query()



"""
绘图类，用于和tensorboard配合
"""
from collections import defaultdict, OrderedDict

from itertools import chain, cycle
from pyecharts import charts
from pyecharts import options as opts
from typing import List, Dict, Union
import numpy as np
from numbers import Number


class Chart:
    def echarts(self):
        raise NotImplementedError

    def matplotlib(self):
        raise NotImplementedError


class Curve(Chart):

    def __init__(self, curve_values: Dict[str, Dict[str, Union[list, str]]], title="-", x_key='x', y_key='y',
                 name_key='name') -> None:
        super().__init__()
        self.title = title
        self.curve_values = curve_values
        self.x_key = x_key
        self.y_key = y_key
        self.name_key = name_key
        self._xaxis = None
        self._aligned = False

    @property
    def x_axis(self):
        if self._xaxis is None:
            self._xaxis = sorted(set(chain(*[v[self.x_key] for v in self.curve_values.values()])))
        return self._xaxis

    def _echarts_align_data(self):
        if self._aligned:
            return
        for k in self.curve_values.keys():
            v = self.curve_values[k]
            x_map = {x: y for x, y in zip(v[self.x_key], v[self.y_key])}
            new_ys = []
            for x in self.x_axis:
                if x in x_map:
                    new_ys.append(x_map[x])
                else:
                    new_ys.append(None)
            v[self.y_key] = new_ys

        self._aligned = True

    def echarts(self):
        self._echarts_align_data()
        c = charts.Line()

        for k, v in self.curve_values.items():
            c.add_xaxis([str(i) for i in self.x_axis])
            break

        # max_v_dict = defaultdict(list)

        for _, v in self.curve_values.items():
            # max_v_dict[k].append([max(v[self.y_key]),min(v[self.y_key])])
            c.add_yaxis(
                v[self.name_key],
                v[self.y_key],
                label_opts=opts.LabelOpts(is_show=False),
                is_symbol_show=False,
            )

        c.set_global_opts(title_opts=opts.TitleOpts(
            title=self.title, padding=5),
            yaxis_opts=opts.AxisOpts(
                min_='dataMin',
                max_='dataMax',
            ),
            legend_opts=opts.LegendOpts(type_='scroll', pos_bottom=10),
            tooltip_opts=opts.TooltipOpts(is_show=True, trigger='axis'), )

        return c

    def matplotlib(self):
        from matplotlib import pyplot as plt
        axes = plt.axes()

        names = []

        for v in sorted(self.curve_values.values(), key=lambda x: x[self.name_key]):
            # for _, v in self.curve_values.items():
            plt.plot(v[self.x_key], v[self.y_key])
            names.append(v[self.name_key])

        plt.legend(names)
        plt.title(self.title)
        return axes


class Parallel(Chart):
    def __init__(self, names: list, tag_dict: Dict[str, List[Number]]):
        super().__init__()
        self.names = names
        self.tag_dict = tag_dict

    def _echarts_axis_opts(self):
        """
        构建坐标系，按照值是string类或数字类，构建相应的坐标系
        Returns:

        """
        res = []
        for i, (k, v) in enumerate(self.tag_dict.items()):
            if isinstance(v[0], str):
                vv = list(set(v))
                vv.sort()
                res.append(opts.ParallelAxisOpts(dim=i, name=k, type_="category", data=vv))
            else:
                res.append(opts.ParallelAxisOpts(dim=i, name=k, max_=max(v), min_=min(v)))

        return res

    def _echarts_data(self):
        """
        返回 echarts需要的数据格式
        Returns:

        """
        res = []
        for i, test_name in enumerate(self.names):
            values = [values[i] for tag, values in self.tag_dict.items()]
            res.append([test_name, values])

        return res

    def echarts(self):
        c = charts.Parallel().add_schema(self._echarts_axis_opts())
        for name, data in self._echarts_data():
            c.add(name, data, is_smooth=True)

        return c

    def _matplotlib_data(self):
        test_key_vals = defaultdict(list)
        for (_, values) in self.tag_dict.items():
            for name, value in zip(self.names, values):
                test_key_vals[name].append(value)

        min_max = OrderedDict()
        for k, v in self.tag_dict.items():
            if isinstance(v[0], str):
                v_min, v_max = '', sorted(v)[0]
                dv = len(set(v)) + 1
            else:
                v_min, v_max = np.min(v), np.max(v)

                dv = v_max - v_min
                if dv == 0:
                    dv = abs(v_min)

                v_min -= dv * 0.05  # add 5% padding below and above
                v_max += dv * 0.05
                dv = v_max - v_min
            min_max[k] = [v_min, v_max, dv]

        return test_key_vals, min_max

    def matplotlib(self):
        from matplotlib import pyplot as plt
        from matplotlib.path import Path
        import matplotlib.patches as patches
        from itertools import cycle

        test_key_vals, min_max = self._matplotlib_data()
        key_size = len(min_max)
        host = plt.axes()
        # img, host = plt.subplots(figsize=(20,10))

        axes = [host] + [host.twinx() for _ in range(key_size - 1)]
        for i, ((k, v), ax) in enumerate(zip(min_max.items(), axes)):
            ax.set_ylim(v[0], v[1])
            if isinstance(v[0], str):
                ax.set_yticks([v[0], *sorted(list(set(self.tag_dict[k]))), '_'])

            ax.spines['top'].set_visible(False)
            ax.spines['bottom'].set_visible(False)
            if ax != host:
                ax.spines['left'].set_visible(False)
                ax.yaxis.set_ticks_position('right')
                ax.spines["right"].set_position(("axes", i / (key_size - 1)))

        host.set_xlim(0, key_size - 1)
        host.set_xticks(range(key_size))
        host.set_xticklabels(list(self.tag_dict.keys()), fontsize=14)
        host.tick_params(axis='x', which='major', pad=7)
        host.spines['right'].set_visible(False)
        host.xaxis.tick_top()
        # host.set_title('Parallel Coordinates Plot', fontsize=18)

        colors = cycle(plt.cm.tab10.colors)
        for test, values in test_key_vals.items():
            # to just draw straight lines between the axes:
            # host.plot(range(ys.shape[1]), zs[j,:], c=colors[(category[j] - 1) % len(colors) ])

            # create bezier curves
            # for each axis, there will a control vertex at the point itself, one at 1/3rd towards the previous and one
            #   at one third towards the next axis; the first and last axis have one less control vertex
            # x-coordinate of the control vertices: at each integer (for the axes) and two inbetween
            # y-coordinate: repeat every point three times, except the first and last only twice
            values = list(values)
            dv0 = None
            v_min0 = None
            for i, (k, (v_min, _, dv)) in enumerate(min_max.items()):
                if isinstance(v_min, str):
                    v_min = 0
                    values[i] = sorted(list(set(self.tag_dict[k]))).index(values[i]) + 1

                if i == 0:
                    dv0 = dv
                    v_min0 = v_min
                    continue
                values[i] = (values[i] - v_min) / dv * dv0 + v_min0

            n = 3
            verts = list(zip([x for x in np.linspace(0, key_size - 1, key_size * n - 2, endpoint=True)],
                             np.repeat(values, n)[1:-1]))
            # verts = list(zip(range(key_size), values))
            # verts = list(zip([0,1,1.5], [2.003,2.004,2.003]))
            # for x,y in verts: host.plot(x, y, 'go') # to show the control points of the beziers
            codes = [Path.MOVETO] + [Path.CURVE4 for _ in range(len(verts) - 1)]
            path = Path(verts, codes)
            patch = patches.PathPatch(path, facecolor='none', lw=1, edgecolor=next(colors))
            host.add_patch(patch)

        host.legend(self.names, ncol=2, bbox_to_anchor=(0.75, -0.05))
        return host


class Table:
    pass


class Bar(Chart):
    pass



"""

"""
import warnings

class BoundCheckError(BaseException):
    pass


class NewParamWarning(Warning):
    pass

class NoneWarning(Warning):
    pass

class AttrTypeNotFoundWarning(Warning):
    pass

warnings.simplefilter('always',NoneWarning)
warnings.simplefilter('always',AttrTypeNotFoundWarning)


"""
Used for recording data
"""
from collections import OrderedDict
from numbers import Number
from typing import Any, Iterable

import numpy as np
import torch

from ..base_classes.trickitems import AvgItem, NoneItem


class Meter:
    """

    Examples:
    --------
    m = Meter()
    m.short()
    m.int()
    m.float()
    m.percent()

    m.k += 10
    m.k.update()

    m.k.backstep()
    """

    def __init__(self):
        self._meter_dict = OrderedDict()
        self._format_dict = dict()
        self._convert_type = []

    def int(self, item: str):
        self._format_dict[item] = lambda x: "{:.0f}".format(x)

    def float(self, key: str, acc=4):
        """
        设置浮点数精度，默认为小数点后四位
        Args:
            key:
            acc:

        Returns:

        """
        self._format_dict[key] = lambda x: "{{:.{}f}}".format(acc).format(x)

    def percent(self, key: str, acc=2):
        self._format_dict[key] = lambda x: "{{:.{}%}}".format(acc).format(x)

    def tensorfloat(self, key: str, acc=4):
        def func(x: torch.Tensor):
            if len(x.shape) == 0:
                return "{{:.{}f}}".format(acc).format(x)
            else:
                return "{{:.{}f}}".format(acc).format(x.item())

        self._format_dict[key] = func

    def str_in_line(self, key: str):
        def func(x):
            l = str(x).split("\n")
            if len(l) > 1:
                return "{}...".format(l[0])
            else:
                return l[0]

        self._format_dict[key] = func

    def add_format_type(self, type, func):
        self._convert_type.append((type, func))

    def _convert(self, val):
        if type(val) in {int, float, bool, str}:
            return val
        # elif isinstance(val, torch.Tensor):
        #     if len(val.shape) == 1 and val.shape[0] == 1:
        #         return val[0]
        for tp, func in self._convert_type:
            if type(val) == tp:
                val = func(val)
        return val

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("_"):
            super().__setattr__(name, value)
        elif name.endswith("_"):
            assert False
        else:
            value = self._convert(value)

            if isinstance(value, int) and name not in self._format_dict:
                self.int(name)
            elif isinstance(value, float) and name not in self._format_dict:
                self.float(name)
            elif isinstance(value, torch.Tensor) and name not in self._format_dict:
                if len(value.shape) == 0 or (sum(value.shape) == 1):
                    self.tensorfloat(name)
                else:
                    self.str_in_line(name)

            self._meter_dict[name] = value

    def __setitem__(self, key, value):
        key = str(key)
        self.__setattr__(key, value)

    def __getattr__(self, item):
        if item.endswith("_"):
            return item.rstrip("_")

        if item not in self._meter_dict:
            return NoneItem()
        else:
            return self._meter_dict[item]

    def items(self):
        for k, v in self._meter_dict.items():
            yield k, v

    def array_items(self):
        """任何数字类型的对象"""
        for k, v in self._meter_dict.items():
            if isinstance(v, (Number, torch.Tensor, np.ndarray)):
                yield k, v

    def numeral_items(self):
        """纯可被记录的数字"""
        for k, v in self._meter_dict.items():
            if isinstance(v, (int, float)):
                yield k, v
            elif isinstance(v, torch.Tensor):
                try:
                    yield k, v.detach().cpu().item()
                except:
                    continue
            elif isinstance(v, np.ndarray):
                try:
                    yield k, v.item()
                except:
                    continue

    def __iter__(self):
        return iter(self._meter_dict)

    def serialize(self):
        """格式化字典"""
        log_dict = OrderedDict()
        for k, v in self._meter_dict.items():
            if k in self._format_dict:
                v = self._format_dict[k](v)
            log_dict[k] = v
        return log_dict

    def __getitem__(self, item):
        item = str(item)
        return self.__getattr__(item)

    def __repr__(self):
        log_dict = self.serialize()

        return " | ".join(["{}: {}".format(k, v) for k, v in log_dict.items()])

    def update(self, meter):
        for k, v in meter.items():
            self[k] = v
        return self

    def map_update(self, items: Iterable, names: Iterable):
        items, names = list(items), list(names)
        assert len(items) == len(names), 'length of items and names not match'
        for item, name in zip(items, names):
            self[name] = item
        return self


class AvgMeter(Meter):
    """

    """

    def __init__(self):
        super().__init__()

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("_"):
            super().__setattr__(name, value)
        elif name.endswith("_"):
            assert False
        else:
            value = self._convert(value)
            if isinstance(value, int) and name not in self._format_dict:
                self.int(name)
            elif isinstance(value, float) and name not in self._format_dict:
                self.float(name)

            if name in self._meter_dict and isinstance(self._meter_dict[name], AvgItem):
                self._meter_dict[name].update(value)
            elif isinstance(value, (float)):
                if name not in self._meter_dict:
                    self._meter_dict[name] = AvgItem()
                self._meter_dict[name].update(value)
            elif isinstance(value, (torch.Tensor, np.ndarray)):
                if isinstance(value, torch.Tensor):
                    value = value.detach().cpu().numpy()
                if len(value.shape) == 0 or sum(value.shape) == 1:
                    if name not in self._meter_dict:
                        self._meter_dict[name] = AvgItem()
                    self._meter_dict[name].update(value)
                else:
                    self._meter_dict[name] = value
            else:
                self._meter_dict[name] = value

    def __getitem__(self, item) -> AvgItem:
        return super().__getitem__(item)

    # def __delitem__(self, key):

    def update(self, meter):
        if meter is None:
            return
        for k, v in meter._meter_dict.items():
            self[k] = v
        self._format_dict.update(meter._format_dict)
        self._convert_type.extend(meter._convert_type)

    def average(self, k):
        self[k] = AvgItem()

    def serialize(self):
        log_dict = OrderedDict()
        for k, v in self._meter_dict.items():
            if k in self._format_dict:
                v = self._format_dict[k](v)
            log_dict[k] = v
        return log_dict

    def array_items(self):
        for k, v in self._meter_dict.items():
            if isinstance(v, (int, float, torch.Tensor, np.ndarray)):
                yield k, v
            elif isinstance(v, AvgItem):
                yield k, v.avg

    def numeral_items(self):
        """"""
        for k, v in self._meter_dict.items():
            if isinstance(v, (int, float)):
                yield k, v
            elif isinstance(v, torch.Tensor):
                try:
                    yield k, v.detach().cpu().item()
                except:
                    continue
            elif isinstance(v, np.ndarray):
                try:
                    yield k, v.item()
                except:
                    continue
            elif isinstance(v, AvgItem):
                yield k, v.avg

    @property
    def meter(self):
        """try to get newest value"""
        return _InMeter(self)

    @property
    def avg(self):
        """try to get average value"""
        return _InMeter(self, mode='avg')


class _InMeter():
    def __init__(self, avgmeter: AvgMeter, mode='newest'):
        self._avg = avgmeter
        self._mode = mode

    def __getattr__(self, item):
        if item not in self._avg._meter_dict:
            raise AttributeError(item)
        res = self._avg._meter_dict[item]
        if isinstance(res, AvgItem):
            if self._mode == 'newest':
                res = res._item
            else:
                res = res.avg

        return res

    def __getitem__(self, item: str):
        return self.__getattr__(item)
